# Copyright (c) 1991 - 1900.  The Regents of the University of
# California (Regents).  All Rights Reserved.
# H+
# 	Title	:  dfconvolve.x
# 	Author	:  tnk
# 	Date	:  04/28/94
# 	Synopsis:  
# 	Keywords:
# 		@(#) dfconvolve.x -r1.1 04/28/94 UCB SSL CEA EGOCS
# 	Revisions:
# 	mm/dd/yy   name		description
# 	04/28/94   tnk	date and time created 94/04/28 12:01:57 by tnk
# H-
# 
#

include <dfset.h>

# @(#)dfconvolve.x	1.1 4/28/94 UCB CEA EGO

# convolve a signal df with a kernal df.
#
# for efficiency, takes the signal df, stuffs it into a zero
# padded work array in memory, and does the convolution outside of the
# df library.
#
# the work array is zero padded to the length of the convolution
# kernal on each end, so that we easily handle convolution end effects
#
#  ---------------------------------------------
# | zero-pad |      signal          | zero-pad |   <-- "work" buffer
#  ---------------------------------------------
# |<-----------worklen ----------------------->|
# |<-kernal->|
#      | <-- convolution --------------->|
#
#
# result is placed back into the incoming signal df

procedure df_convolve(signal_df,x_spacing,kernal_df)
pointer signal_df         #I/O DF signal to convolve on (PD)
double  x_spacing         #I spacing to which to map signal & kernal
pointer kernal_df         #I kernal (PF)

int     siglen   # length of signal array 
int     kernlen  # length of kernal array
int     worklen  # length of work array
pointer kern     # kernal array
pointer work_x   # working array (x)
pointer work_y   # working array (y)
pointer out_y    #
pointer sp       # stack pointer
double  xmin, xmax, ymin, ymax
int     n_sig_steps
int     n_pad_steps
int     n_tot_steps
double  x0
int     nread
pointer new_sig
pointer new_kern

# FUNCTIONS
pointer df_open()
int     df_len()
int     df_dgetd()

begin

   call smark(sp)

   # make kernal regularly spaced, on x_spacing widths
   new_kern = df_open(DF_PF,TY_DOUBLE,TY_DOUBLE)
   call df_pedge(new_kern,0.0d0,0.0d0)

   call df_minmax(kernal_df,xmin,xmax,ymin,ymax)
   n_sig_steps=(xmax-xmin)/x_spacing+1
   x0 = xmin 
   n_tot_steps = n_sig_steps + 1 
   call df_rsdomain(kernal_df,x0,x_spacing,n_tot_steps,new_kern)
   kernlen = df_len(new_kern)

   new_sig = df_open(DF_PD,TY_DOUBLE,TY_DOUBLE)
   call df_pedge(new_sig,0.0d0,0.0d0)

   # change the signal domain so that it's evenly spaced
   # (incidentally, this also pads with zeros on the right & left
   # assuming the df edges are zero)
   call df_minmax(signal_df,xmin,xmax,ymin,ymax)
   n_sig_steps=(xmax-xmin)/x_spacing+1
   n_pad_steps=kernlen
   x0 = xmin - (x_spacing * n_pad_steps)
   n_tot_steps = n_sig_steps + n_pad_steps*2 + 1 
   call df_rsdomain(signal_df,x0,x_spacing,n_tot_steps,new_sig)
   siglen = df_len(new_sig)

   worklen = n_tot_steps
   call salloc(work_x,worklen,TY_DOUBLE)
   call salloc(work_y,worklen,TY_DOUBLE)
   call salloc(out_y, worklen,TY_DOUBLE)  
   call salloc(kern,  kernlen,TY_DOUBLE)

   # zero buffers
   call aclrd(Memd[work_x],worklen)
   call aclrd(Memd[work_y],worklen)
   call aclrd(Memd[out_y], worklen)
   call aclrd(Memd[kern],  kernlen)

   # convert DFs to arrays
   nread = df_dgetd(new_kern,Memd[work_x],Memd[kern],kernlen)
   nread = df_dgetd(new_sig,Memd[work_x],Memd[work_y],worklen)

   call do_conv(Memd[work_x],Memd[work_y],worklen,
     Memd[out_y],Memd[kern],kernlen)

   # transfer convolution back into signal DF 
   call df_reset(signal_df)
   call df_dputd(signal_df,Memd[work_x],Memd[out_y],worklen)
   call df_pedge(signal_df,0.0d0,0.0d0)
   call df_px0(signal_df,Memd[work_x]-x_spacing)

   call df_close(new_kern)
   call df_close(new_sig)

   call sfree(sp)

end

# do a kernal convolution on data in arrays
procedure do_conv(work_x,work_y,worklen,out_y,kern,kernlen)
double  work_x[ARB]  #I "work buffer" x values
double  work_y[ARB]  #I work buffer y values (contains signal & padding)
int     worklen      #I length of work buffer
double  out_y[ARB]   #O output buffer (length=worklen)
double  kern[ARB]    #I kernal buffer
int     kernlen      #I length of kern buffer

int     start
int     half_kernlen
double  y_convolve
int     i
int     k
int     j
double  old_x
double  new_y

begin

   # assumes signal is zero padded on both ends (kernlen zeros)
   # and that both kernal & signal are similarly spaced.

   # for each bin
   #   convolve the kernal with the signal (ie, multiply & integrate)
   #   save off the new point 

   # note that we start convolving halfway into the kernal-length padding

   # convolve 
   half_kernlen = kernlen/2
   start = half_kernlen
   for (i=start; i<worklen-half_kernlen; i=i+1) {

      # multiply signal by kernal and integrate
      k = 1
      j = i+k-1-half_kernlen
      old_x = work_x[j]
      y_convolve = 0.0d0
      for (k=2; k<=kernlen; k=k+1) {
         j = i+k-1-half_kernlen
         new_y = kern[k] * work_y[j]
         y_convolve = y_convolve + new_y * (work_x[j] - old_x)
         old_x = work_x[j]
      }
      out_y[i] = y_convolve
   }

end
