# Copyright (c) 1991 - 1900.  The Regents of the University of
# California (Regents).  All Rights Reserved.
# H+
# 	Title	:  erquat.x
# 	Author	:  ericco
# 	Date	:  05/07/91
# 	Synopsis:  
# 	Keywords:
# 		@(#) erquat.x -r1.10 05/07/91 UCB SSL CEA EGOCS
# 	Revisions:
# 	mm/dd/yy   name		description
# 	07/16/93   akeith	clearified er_rotate
# 	05/05/93   akeith	comment
# 	05/05/93   ericco	added er_qmkrotate
# 	05/04/93   akeith	fixed theta == 0
# 	09/28/92   akeith	put in debug statements
# 	07/17/92   ericco	fixed orientation error in er_qfrompos
# 	07/16/92   ericco	fixed rotation operations
# 	10/29/91   ericco	added eu_malloc. also fixed point mode
# 	05/07/91   ericco	date and time created 91/05/07 14:11:03 by ericco
# H-
# 
#

#	Revised: 7/16/93

#	Routines to manipulate quaternions

#	A quaternion is a four vector consisting of a rotation angle
#	and a rotation axis.
#	<W, X, Y, Z> :=
#		<cos (theta), x sin(theta), y sin(theta), z sin(theta)>
#	where theta is the half the rotation, and <x, y, z> is the
#	rotation axis.
#	W^2 + X^2 + Y^2 + Z^2 = 1.
#	W range from -1 to 1.
#
#	From "Quaternion calculus as a basic tool in computer graphics",
#	by Daniel Pletinckx, in Visual Computer, 1989 5:2-13.

include	<math.h>
include <mach.h>


define	W	1
define	X	2
define	Y	3
define	Z	4


#	sin(angle) ~ angle
define	SIN_EPSILON	1.0e-20
define	NEAR_ONE	0.9999999
define	NEAR_ZERO	1.0e-07
define	PI		3.141592654


#	PROCEDURE er_qcompose creates a quaternion from an unnormalized rotation
#	axis, and an unnormalized rotation.
	procedure er_qcompose (x, y, z, theta, qt)
double	x, y, z				#I	rotation axis vector
double	theta				#I	rotation radians
double	qt[4]				#O	new quaternion

double	theta2				#L	half rotation angle
double	stheta2				#L	sin(theta2)
double	norm				#L	axis normalization factor
double	sin(), cos(), sqrt()		#L	local functions
begin
	# normalize theta
	if (theta > TWOPI) {
		repeat
			theta = theta - TWOPI
		until (theta <= TWOPI)
	} else if (theta < -TWOPI) {
		repeat
			theta = theta + TWOPI
		until (theta >= -TWOPI)
	}
	norm = sqrt (x * x + y * y + z * z)
	if (norm < NEAR_ZERO) {		#	error condition ?
					#	this is no rotation
		qt[W] = 1
		qt[X] = 0
		qt[Y] = 0
		qt[Z] = 0
	} else {			#	good input conversion
		norm = 1.0 / norm
		theta2 = theta / 2.0

		qt[W] = cos (theta2)
		stheta2 = sin (theta2)
		qt[X] = stheta2 * x * norm
		qt[Y] = stheta2 * y * norm
		qt[Z] = stheta2 * z * norm
	}
end


#	Procedure er_qdecompose converts a quaternion into a rotation axis and
#	a rotation. The axis is not normalized, -TWOPI <= theta <= TWOPI.
	procedure er_qdecompose (qt, x, y, z, theta)
double	qt[4]				#I	quaternion
double	x, y, z				#O	rotation axis vector
double	theta				#O	rotation radians

double	theta2				#L	half rotation angle
double	istheta2			#L	inverse sin(theta2)
double	acos(), sin()			#L	local functions
begin
	if (qt[W] > 1.0) {		# round off error ?
		theta = 0
		x = 1
		y = 0
		z = 0
	} else if (qt[W] < -1.0) {	# round off error ?
		theta = TWOPI
		x = 1
		y = 0
		z = 0
	} else {			# nominal case
		theta2 = acos (qt[W])
		theta = 2 * theta2
		if (theta2 < EPSILOND) {	# any axis will do
			x = 1
			y = 0
			z = 0
		} else {
			istheta2 = 1.0 / sin(theta2)
			x = istheta2 * qt[X]
			y = istheta2 * qt[Y]
			z = istheta2 * qt[Z]
		}
	}
end


#	PROCEDURE er_qmult computes the composition of q1
#	and q2 such that q = q1*q2. q = q1*q2 represents the results
#	of successive rotations first by q1 and then by q2.
#	result q must be distinct from inputs q1 and q2
	procedure er_qmult (q1, q2, q)
double	q1[4], q2[4]			#I	quaternions
double	q[4]				#O	quaternion

begin
	q[W] = q1[W] * q2[W] - q1[X] * q2[X] - q1[Y] * q2[Y] - q1[Z] * q2[Z]
	q[X] = q1[X] * q2[W] + q1[W] * q2[X] - q1[Z] * q2[Y] + q1[Y] * q2[Z]
	q[Y] = q1[Y] * q2[W] + q1[Z] * q2[X] + q1[W] * q2[Y] - q1[X] * q2[Z]
	q[Z] = q1[Z] * q2[W] - q1[Y] * q2[X] + q1[X] * q2[Y] + q1[W] * q2[Z]
end


#	PROCEDURE er_qinvert computes the inverse of a quaternion.
#	result q can be input q1
	procedure er_qinvert (q1, q)
double	q1[4]				#I	quaternion to invert
double	q[4]				#O	inverted quaternion

begin
	q[W] = q1[W]
	q[X] = -q1[X]
	q[Y] = -q1[Y]
	q[Z] = -q1[Z]
end



#	PROCEDURE er_qscale multiplies the angular value of a quaternion.
#	result q can be input q1
	procedure er_qscale (q1, scalar, q)
double	q1[4]				#I	quaternion
double	scalar				#I	scalar
double	q[4]				#O	scaled quaternion

double	x, y, z				#L	rotation axis
double	theta				#L	rotation angle
begin
	call er_qdecompose (q1, x, y, z, theta)
	theta = scalar * theta
	call er_qcompose (x, y, z, theta, q)
#	assertq(q,"er_qscale","")
end


#	PROCEDURE er_qdiv computes q3 = q2' * q1 where q2' is 
#	the conjugate of q2. q3 is the quaternion when composed
#	with q2 is q1 = q2 * q3.
#	result q3 can be input q2, but not input q1
	procedure er_qdiv (q1, q2, q3)
double	q1[4]				#I	initial quaternion
double	q2[4]				#I	final quaternion
double	q3[4]				#O	delta quaternion

double	qt[4]
begin
	call er_qinvert (q2, qt)
	call er_qmult (qt, q1, q3)
end


#	PROCEDURE er_qextrapolate returns quaternion q which is on the projection
#	from q1 to q2 at a distance scalar, where scalar = 0 implies q1 and
#	scalar = 1 implies q2.  This routine has only been tested with scalar
#	between 0 and 1.
#	result q can be input q1, but not input q2
	procedure er_qextrapolate (q1, q2, scalar, q)
double	q1[4], q2[4]			#I	input quaternions
double	scalar				#I	any scalar
double	q[4]				#O	output quaternion

double	qt[4]				#L	temp space
begin
	call er_qdiv (q2, q1, qt)
	call er_qscale (qt, scalar, qt)
	call er_qmult (q1, qt, q)
end


#	PROCEDURE er_qinterpolate returns quaternion q which is on the projection
#	from q1 to q2 at a distance scalar, where scalar = 0 implies q1 and
#	scalar = 1 implies q2.  This routine has only been tested with scalar
#	between 0 and 1. If sum is NEAR_ONE, then q1 and q2 are nearly equil.
#	result q can not be q1 or q2.
	procedure er_qinterpolate (q1, q2, scalar, q)
double	q1[4], q2[4]			#I	input quaternions
double	scalar				#I	0 < scalar < 1
double	q[4]				#O	output quaternion

double	sum, beta1, beta2, theta, istheta
double	acos(), sin()
begin
	sum = q1[W] * q2[W] + q1[X] * q2[X] + q1[Y] * q2[Y] + q1[Z] * q2[Z]
	if (sum >= NEAR_ONE || sum <= -NEAR_ONE) {
		beta1 = 1 - scalar
		beta2 = scalar
	} else {
		theta = acos (sum)
		istheta = 1.0 / sin (theta)
		beta1 = sin ((1 - scalar) * theta) * istheta
		beta2 = sin (scalar * theta) * istheta
	}

	q[W] = beta1 * q1[W] + beta2 * q2[W]
	q[X] = beta1 * q1[X] + beta2 * q2[X]
	q[Y] = beta1 * q1[Y] + beta2 * q2[Y]
	q[Z] = beta1 * q1[Z] + beta2 * q2[Z]
end


#	PROCEDURE er_qcheck returns TRUE iff the quaternion is normalized.
bool	procedure er_qcheck (q)
double	q[4]

double	norm
begin
	if (q[W] > 1.0)
		return FALSE
	if (q[W] < -1.0)
		return FALSE
	norm = q[W] * q[W] + q[X] * q[X] + q[Y] * q[Y] + q[Z] * q[Z] - 1
	if (norm < -EPSILON)
		return FALSE
	if (norm > EPSILON)
		return FALSE

	return TRUE
end

#	PROCEDURE er_qfrompos converts a direction into a quaternion.
	procedure er_qfrompos (ra, dec, q)
double	ra, dec				#I	direction
double	q[4]				#O	vector quaternion

double	cos_dec				#L	cosine of declination
double	cos(), sin()			#L	local functions
begin
	q[W] = 0			# just because
	cos_dec = cos(dec)
	q[X] = cos_dec * cos(ra)	# just geometry
	q[Y] = cos_dec * sin(ra)
	q[Z] = sin(dec)
end


#	PROCEDURE er_qfromvec
	procedure er_qfromvec (x, y, z, q)
double	x, y, z				#I	any vector
double	q[4]				#O	vector quaternion

begin
	q[W] = 0
	q[X] = x
	q[Y] = y
	q[Z] = z
end

#	PROCEDURE er_qtovec
	procedure er_qtovec (q, x, y, z)
double	q[4]				#I	vector quaternion
double	x, y, z				#O	any vector

begin
	x = q[X]
	y = q[Y]
	z = q[Z]
end

#	PROCEDURE er_qtopos converts a "vector" quaternion into a
#	direction.  the quaternion is not checked.
	procedure er_qtopos (v, ra, dec)
double	v[4]				#I	vector quaternion
double	ra, dec				#O	direction

double	atan2(), sqrt()
begin
	if (v[X] == 0 && v[Y] == 0) {
		ra = 0
		if (v[Z] > 0)
			dec = HALFPI
		else
			dec = -HALFPI
	} else {
		dec = atan2 (v[Z], sqrt(v[X]**2 + v[Y]**2))
		ra = atan2(v[Y], v[X])
	}
end


# v' = qvq' where q' is q transposed
# but vw = 0
#    = [0, (w1^2 - 1)v2 + 2w1(v1xv2) + 2(v2.v1)v1] <-- not sure
# this could be used to speed up this computation
# this could all be a function of w1, v1, ra, and dec into
# ra and dec.
#
#	PROCEDURE er_qrotateq applies a quaternion rotation to a
#	quaternion v' = q_star * v * q  where q_star is the conjugate
#	of q. v' is the representaion of v in the primed coordinate
#	system. WARNING: this implies a rotation of the coordinates.
#
	procedure er_qrotateq (q, r)
double	q[4]
double	r[4]

double	t[4], tmp[4]
begin
	call er_qinvert (r, t)
	call er_qmult (t, q, tmp)
	call er_qmult (tmp, r, q)
end

#	PROCEDURE er_qmkrotate creates the smallest rotation quaternion
#	such that b = q_star * a * q  where q_star is the conjugate of q.
	procedure er_qmkrotate (q, a, b)
double	q[4]
double	a[4]
double	b[4]

double	theta
double	t[4]
double	acos()

begin
	call er_qmult (b, a, t)
	theta = acos(-t[W])
	call er_qcompose(t[X], t[Y], t[Z], theta, q)
end

#	PROCEDURE er_qrotate applies a quaternion rotation on a direction
	procedure er_qrotate (ra, dec, q)
double	ra, dec				#B	position in radians
double	q				#I	quaternion rotation

double	v[4]				#L	position vector
begin
	call er_qfrompos (ra, dec, v)
	call er_qrotateq (v, q)
	call er_qtopos (v, ra, dec)
end


#	PROCEDURE er_qnormalize normalizes the quaternion
	procedure er_qnormalize (q1, q)
double	q1[4]				#I	unnormalize quaternion
double	q[4]				#O	normalized quaternion

double	norm
double	sqrt()
begin
	norm = sqrt (q1[W]*q1[W] + q1[X]*q1[X] + q1[Y]*q1[Y] + q1[Z]*q1[Z])
	if (norm < EPSILOND) {
		q[W] = 1
		q[X] = 0
		q[Y] = 0
		q[Z] = 0
	} else {
		norm = 1.0 / norm
		q[W] = q1[W] * norm
		q[X] = q1[X] * norm
		q[Y] = q1[Y] * norm
		q[Z] = q1[Z] * norm
		if (q[W] > 1.0) {
			q[W] = 1
			q[X] = 0
			q[Y] = 0
			q[Z] = 0
		} else if (q[W] < -1.0) {
			q[W] = -1
			q[X] = 0
			q[Y] = 0
			q[Z] = 0
		}
	}

#	assertq(q,"er_qnormalize","")
end


#	PROCEDURE er_qcopy copies a quaternion from q to qt
	procedure er_qcopy (q, qt)
double	q[4]				#I	quaternion
double	qt[4]				#O	quaternion

begin
	qt[W] = q[W]
	qt[X] = q[X]
	qt[Y] = q[Y]
	qt[Z] = q[Z]
end


#	PROCEDURE er_qmalloc returns a pointer to a heap space quaternion.
#	The returned pointer can only be used with Memd.
pointer	procedure er_qmalloc ()

pointer	ret
begin
	call eu_malloc (ret, 4, TY_DOUBLE)
	return ret
end


#	PROCEDURE er_qsalloc returns a pointer to a stack space quaternion.
#	The returned pointer can only be used with Memd.
pointer	procedure er_qsalloc ()

pointer	ret
begin
	call salloc (ret, 4, TY_DOUBLE)
	return ret
end


#	PROCEDURE er_qmfree returns a heap space quaternion to the heap.
	procedure er_qmfree (qptr)
pointer	qptr

begin
	call eu_mfree (qptr, TY_DOUBLE)
end
