# Copyright (c) 1991 - 1900.  The Regents of the University of
# California (Regents).  All Rights Reserved.
# H+
# 	Title	:  rumphism.x
# 	Author	:  tnk
# 	Date	:  08/05/93
# 	Synopsis:  
# 	Keywords:
# 		@(#) rumphism.x -r1.2 08/05/93 UCB SSL CEA EGOCS
# 	Revisions:
# 	mm/dd/yy   name		description
# 	09/27/93   tnk	include the Morrison and McCammon cross sections below 43.65 Ang
# 	08/05/93   tnk	date and time created 93/08/05 15:55:16 by tnk
# H-
# 
#


# @(#)rumphism.x	1.2 9/27/93 UCB CEA EGO
# T. Kilsdonk
# 4 Aug 94

# SPP conversion of Pat Jelinksky's Rumph ISM modeling code (ismatten.c)
# which was last revised 9/24/93.

# Synopsis: containes several subroutines that are useful
#   in the calculation of the attenuation from the interstellar medium.
#
#   "tauhe" calculates the optical depth due to neutral helium
#   "tauh" calculates the optical depth for any hydrogenic atom
#   "rumphism" calculates the transmission of neutral hydrogen,
#	neutral helium, and once ionized helium based on the Rumph et. al.
#       model
#
# Now includes the latest He I cross sections obtained
# from Stefan Vennes, who obtains them from Frits Paerles, who fit the data of 
# Marr and West (1976) "Atomic Data and Nuclear Data Tables, 18, 497".
#
# Now also includes the auto-ionization cross sections of helium
# from Rumpf et al.  - July 24 1992.
#
# Now also include the Morrison and McCammon cross sections below
# 43.65 A. - September 24, 1993

include	<math.h>

#**************************************************************************
# rumphism returns the attenuation due to the Rumph et al. ISM model

double	procedure rumphism(wav,hcol,heicol,heiicol)     #LOCAL
double	wav	# Wavelength (Angstroms) 
double	hcol	# Neutral Hydrogen column Density (cm**-2)
double	heicol	# Neutral Helium column Density (cm**-2)
double	heiicol	# Ionized Helium column Density (cm**-2)

double tau,trans
double tauh()
double tauhe()
double mmcross()

define	MMWAV	43.6549d0

begin

   if(wav < MMWAV) {
	tau = hcol * mmcross(wav) + tauhe(wav,heicol) + 
 			tauh(wav,heiicol,2.0d0);
   }
   else {
	tau = tauh(wav,hcol,1.0d0) + tauh(wav,heiicol,2.0d0) + 
           tauhe(wav,heicol)
   }

   trans = exp(-tau)
   return(trans)

end

#**************************************************************************
#
#  tauh returns the ism optical depth 
#  for any hydrogenic atom as per Spitzer (Physical processes in
#  the interstellar medium)  Page 105.
#
double	procedure tauh(wav,hcol,zee)   #LOCAL
double	wav    # wavelength (angstroms)
double	hcol   # neutral helium column density (cm**2)
double	zee    # number of protons in atom

double	z,sigma,tau,ratio


begin
	
   ratio = zee * zee * wav / 911.75d0
   if (ratio < 1.0d0) {
	z = sqrt(ratio/(1.0d0-ratio))
	sigma = 3.44d-16 * ratio**4.0d0 * exp(-4.0d0 * z * atan(1.0d0/z))/
		((1.0d0 - exp(-6.283185308d0*z))*zee*zee)
	tau = hcol * sigma
   }
   else {
	tau = 0.0d0
   }
   return(tau)

end

#**************************************************************************
#
# tauhe returns the neutral helium optical depth from a log-log polynomial 
# interpolation of the measured cross sections.
#
# cross sections are from experimental data compiled by Marr & West (1976)
# Atomic Data and Nuclear Data Tables, 18, 497
# Code based on fortran code from (Frits Paerels 12/90)
# PatJ got the code from Stefan Vennes (4/1992)
# the inputs are the wavelength, wav, in angstroms, 

double procedure tauhe(lambda,hecol)   #LOCAL
double lambda	#  wavelength in Angstroms
double hecol	#  neutral helium column density (cm**-2) 

double x
double y
double c1[8]
double c2[8]
double fano_q[4]  
double fano_nu[4] 
double fano_gamma[4]
double fano()

begin

   # From experimental data compiled by Marr & West (1976)
   # Atomic Data and Nuclear Data Tables, 18, 497
   # (from Frits Paerels 12/90) (then from Stefan Vennes 4/92)

   # polynomial coeffients for He I cross section to use for wavelengths 
   # greater than or equal to 46 A.

   c1[1] = -2.953607d+01
   c1[2] =  7.083061d+00
   c1[3] =  8.678646d-01
   c1[4] = -1.221932d+00
   c1[5] =  4.052997d-02
   c1[6] =  1.317109d-01
   c1[7] = -3.265795d-02
   c1[8] =  2.500933d-03

   # polynomial coeffients for He I cross section to use for wavelengths 
   # less than 46 A.

   c2[1] = -2.465188d+01
   c2[2] =  4.354679d+00
   c2[3] = -3.553024d+00
   c2[4] =  5.573040d+00
   c2[5] = -5.872938d+00
   c2[6] =  3.720797d+00
   c2[7] = -1.226919d+00
   c2[8] =  1.576657d-01

   # parameters of autoionization resonances for 4 strongest for helium
   # Numbers are from Oza (1986), Phys Rev. A, 33, 824 -- nu and gamma
   # and Fernley et al., J. Phys. B., 20, 6457, 1987 -- q

   fano_q[1] =  2.81d0
   fano_q[2] =  2.51d0
   fano_q[3] =  2.45d0
   fano_q[4] =  2.44d0

   fano_nu[1] =  1.610d0
   fano_nu[2] =  2.795d0
   fano_nu[3] =  3.817d0
   fano_nu[4] =  4.824d0

   fano_gamma[1] = 2.64061d-03
   fano_gamma[2] = 6.20116d-04
   fano_gamma[3] = 2.56061d-04
   fano_gamma[4] = 1.320159d-04



   if(lambda > 503.97d0) {	# if wavelength is above ionization limit 
   	return(0.0d0)	# then there is no absorption	
   }

   x=log10(lambda)	# polynomial fits use log of lambda 

   if(lambda < 46.0d0) {	# if wavelength < 46 use c2 
        y=c2[1]+x*(c2[2]+x*(c2[3]+x*(c2[4]+x*(c2[5]+x*(c2[6]+
                                x*(c2[7]+x*c2[8]))))))
   }
   else {   		# if wavelength > 46.0 use c1 
   	y=c1[1]+x*(c1[2]+x*(c1[3]+x*(c1[4]+x*(c1[5]+x*(c1[6]+
   				x*(c1[7]+x*c1[8]))))))
   	y = y + log10(fano(fano_q[1],fano_nu[1],fano_gamma[1],lambda))
   	y = y + log10(fano(fano_q[2],fano_nu[2],fano_gamma[2],lambda))
   	y = y + log10(fano(fano_q[3],fano_nu[3],fano_gamma[3],lambda))
   	y = y + log10(fano(fano_q[4],fano_nu[4],fano_gamma[4],lambda))
   }
	
   return(hecol * 10.0d0**y)
end

#**************************************************************************
# fano returns a Fano line profile for use by tauhe.
# The form of this line shape is taken from:
#	
#	Fernley, Taylor and Seaton; Journal of Physics (B), 20: 6457 
#       (1987).
#
double procedure fano(q, nu, gamma, lambda)   #LOCAL
double q	
double nu	
double gamma
double lambda

double epsilon
double esubi
double x

begin

   epsilon = 911.2671d0/lambda   # energy in rydbergs 
   esubi = 3.0d0 - 1.0d0/(nu*nu) + 1.807317d0 
   x = 2.0d0*((epsilon-esubi)/gamma)

   return(((x-q)*(x-q))/(1.0d0+x*x))

end

#**************************************************************************
# mmcross finds the Morrison and McCammon cross section minus the
#	helium cross section
#
# Uses the fits from Morrison and McCammon 1983
#   for the ISM cross sections.
#
#
double	procedure mmcross(wav)    #LOCAL
double	wav			# wavelength in Angstroms 

define	nMcCammon  12

int	i
double	sigma, sigmahe
double  minWav[nMcCammon]
double  MaxWav[nMcCammon]
double  mc0[nMcCammon]
double  mc1[nMcCammon]
double  mc2[nMcCammon]

double	mmcrosshe()

begin

   # Coefficients from Morrison and McCammon 1983 with fixed
   #   abundances.

   minWav[1] = 1.2398d0
   minWav[2] = 1.48818d0
   minWav[3] = 1.7435d0
   minWav[4] = 3.07033d0
   minWav[5] = 3.86231d0
   minWav[6] = 5.0174d0
   minWav[7] = 6.73804d0
   minWav[8] = 9.51497d0
   minWav[9] = 14.2999d0
   minWav[10] = 17.5361d0
   minWav[11] = 23.3045d0
   minWav[12] = 30.995d0

   maxWav[1] = 1.48818d0
   maxWav[2] = 1.7435d0
   maxWav[3] = 3.07033d0
   maxWav[4] = 3.86231d0
   maxWav[5] = 5.0174d0
   maxWav[6] = 6.73804d0
   maxWav[7] = 9.51497d0
   maxWav[8] = 14.2999d0
   maxWav[9] = 17.5361d0
   maxWav[10] = 23.3045d0
   maxWav[11] = 30.995d0
   maxWav[12] = 43.6549d0

   mc0[1] = 0.367948d0
   mc0[2] = 0.330062d0
   mc0[3] = 0.227685d0
   mc0[4] = 0.184814d0
   mc0[5] = 0.179829d0
   mc0[6] = 0.106365d0
   mc0[7] = 0.0741459d0
   mc0[8] = 0.0632838d0
   mc0[9] = 0.162093d0
   mc0[10] = 0.0501128d0
   mc0[11] = 0.0374665d0
   mc0[12] = 0.0409823d0

   mc1[1] = 0.163945d0
   mc1[2] = 0.201027d0
   mc1[3] = -0.0156138d0
   mc1[4] = 0.121657d0
   mc1[5] = 0.121657d0
   mc1[6] = 0.681151d0
   mc1[7] = 0.955043d0
   mc1[8] = 1.10142d0
   mc1[9] = -2.47608d0
   mc1[10] = 0.948537d0
   mc1[11] = 0.434583d0
   mc1[12] = 0.122308d0

   mc2[1] = 0d0
   mc2[2] = 0d0
   mc2[3] = 0.0604936d0
   mc2[4] = 0d0
   mc2[5] = 0d0
   mc2[6] = -1.37119d0
   mc2[7] = -2.54073d0
   mc2[8] = -3.84739d0
   mc2[9] = 23.7135d0
   mc2[10] = -4.92821d0
   mc2[11] = -4.14583d0
   mc2[12] = 0.34683d0

   if(wav > maxWav[nMcCammon]) {
	call error(999,"Wavelength too large for Morrison data")
   }

   i = 1
   while(wav > maxWav[i]) {
	i=i+1
   }

   sigma = wav*(mc2[i] + wav*(mc1[i] + wav*mc0[i]))*1.0d-24
   sigmahe = mmcrosshe(wav)
   return(sigma - 0.1d0*sigmahe)

end

#**************************************************************************
#
#	A fit to the helium cross section from 
#	Henke, B. L., Lee, P., Tanaka, T. J., Shimabukuro, R. L., and
#	Fujikawa, B. K. 1982, Atom. and Nucl. Data Tables, 27, 22.
#	Same reference as used in Morrison and McCammon 1983
#									
double procedure mmcrosshe(wav)   # LOCAL
double	wav		# wavelength in Angstroms 

double	x
double	sigma
double	coeff[5]
double	coeff1[2]

double log10()
define	SPLITWAV  8.34d0

begin

   coeff[1] = -25.080222d0
   coeff[2] = 0.30642752d0
   coeff[3] = -0.035003598d0
   coeff[4] = 0.58009643d0
   coeff[5] = -0.47433857

   coeff1[1]= -24.864486d0
   coeff1[2]= 0.16154438d0

   x = log10(wav)

   if(wav <= SPLITWAV) {
	sigma = coeff[1] + x*(coeff[2] + x*(coeff[3] + 
		x*(coeff[4] + x*coeff[5])))
   }
   else {
	sigma = coeff1[1] + x*coeff1[2]
   }

   return( 10.0d0**sigma * (wav*wav*wav))

end
