# Copyright (c) 1991 - 1900.  The Regents of the University of
# California (Regents).  All Rights Reserved.
# H+
# 	Title	:  t_ism.x
# 	Author	:  tnk
# 	Date	:  08/05/93
# 	Synopsis:  
# 	Keywords:
# 		@(#) t_ism.x -r1.1 08/05/93 UCB SSL CEA EGOCS
# 	Revisions:
# 	mm/dd/yy   name		description
# 	08/05/93   tnk	date and time created 93/08/05 15:55:17 by tnk
# H-
# 
#

# @(#)t_ism.x	1.1 8/5/93 UCB CEA EGO
# AUTHOR : T. Kilsdonk
# DATE : 3 Aug 93

# SPP adaptation of Pat Jelinsky's ISM code (ism.c)

# ism adds the interstellar absorption to a spectral flux 
# file. The input ASCII file must have the format of two columns. The
# first column is the wavlength in angstroms, and the
# second column is the flux in any units. The program
# then writes the resultant wavelength & flux for each wavelength 
# to the output file.
#
# The input model parameters are:
#    neutral hydrogen column (cm**-2)
#    heI/HI ratio
#    heII/HI ratio
#
# The only model currently supported is the "Rumph" model.
#
# Documenting cards in the output stream have a '#' in the left column
#
# If either the input wavelength or flux is invalid, the line
# number in the input file & an error message is sent to stderr.

procedure t_ism()    #EXTERNAL

string HCOL_PARM           "hcolumn"
string HEI_H_RATIO_PARM    "he1hratio"
string HEII_H_RATIO_PARM   "he2hratio"
string MODEL_PARM          "model"
define INNAME_PARM         "input"       # output filename (STDIN)
define OUTNAME_PARM        "output"      # output filename (STDOUT)
string DOCUMENT_PARM       "document"

define MODEL_DICT          "|rumph|"     # dictionary of model names
define RUMPH_INDEX         1             # index of model in MODEL_DICT

double  hcol           # hydrogen column density
double  ratio          # HeI/HI ratio
double  ratio1         # HeII/HI ratio
double  heicol         # HeI column density
double  heiicol        # HeII column density
double  wav            # wavelength
double  flux           # flux
double  ismflux        # ism attenuated flux
int     in_fd          # output file descriptor (ASCII)
int     out_fd         # output file descriptor (ASCII)
pointer in_name        # output file name (stdin is ok)
pointer out_name       # output file name (stdout is ok)
pointer model          # name of ISM model
bool    document       # boolean: include document cards?
pointer temp_str       # temp string
pointer stackp         # stack pointer
int     linenum        # input file line number
int     model_index    # index number of model in model dict list

# Functions
double rumphism()     # rumph ism model
double clgetd()       #
bool   clgetb()       #
int    fscan()        #
int    open()         # open file
bool   streq()        # string compare
int    strdic()       # dictionary lookup

begin

   # mark the stack
   call smark(stackp)              

   call salloc(model,    SZ_LINE, TY_CHAR)
   call salloc(in_name,  SZ_LINE, TY_CHAR)
   call salloc(out_name, SZ_LINE, TY_CHAR)
   call salloc(temp_str, SZ_LINE, TY_CHAR)

   # get CL parameters 
   hcol   =   clgetd(HCOL_PARM)           # neutral hydrogen column (cm**-2)
   ratio  =   clgetd(HEI_H_RATIO_PARM)    # heI/HI ratio
   ratio1 =   clgetd(HEII_H_RATIO_PARM)   # heII/Hi ratio
   call       clgwrd(MODEL_PARM, Memc[model], SZ_LINE, MODEL_DICT) #model name
   call       clgstr(INNAME_PARM,  Memc[in_name], SZ_LINE)
   call       clgstr(OUTNAME_PARM, Memc[out_name], SZ_LINE)
   document = clgetb(DOCUMENT_PARM)     # put out documenting cards

   # check for stdin (make case insensitive)
   call strcpy(Memc[in_name],Memc[temp_str],SZ_LINE)     
   call strlwr(Memc[temp_str])

   if (streq(Memc[temp_str],"stdin") )
      in_fd = STDIN
   else {
      in_fd = open(Memc[in_name], READ_ONLY, TEXT_FILE)    
   }

   # check for stdout (make case insensitive)
   call strcpy(Memc[out_name],Memc[temp_str],SZ_LINE)     
   call strlwr(Memc[temp_str])

   if (streq(Memc[temp_str],"stdout") )
      out_fd = STDOUT
   else {
      out_fd = open(Memc[out_name], NEW_FILE, TEXT_FILE)    

   }

   # put out "documenting" cards
   if (document) {

      call fprintf(out_fd,"# neutral hydrogen column density (cm**-2): %10.3e\n")
      call pargd(hcol)

      call fprintf(out_fd,"# HeI/HI ratio:  %.2f\n")
      call pargd(ratio)

      call fprintf(out_fd,"# HeII/HI ratio: %.2f\n")
      call pargd(ratio1)

      call fprintf(out_fd,"# ISM model:     %s\n")
      call pargstr(Memc[model])

   }

   # use ratios to find other column densities
   heicol = hcol * ratio;
   heiicol = hcol * ratio1;

   # figure out which model was selected
   model_index = strdic(Memc[model],Memc[temp_str],SZ_LINE,MODEL_DICT)

   # go through each wavelength, flux pair and put out 
   # wavelength and attenuated flux

   linenum = 0
   while( fscan(in_fd) != EOF) {

      linenum = linenum + 1

      wav=-1.0
      flux=-1.0

      call gargd(wav)
      call gargd(flux)

      if (wav < 0.0 || flux < 0.0) {
         call fprintf(STDERR,"line %d: invalid wavelength, flux pair\n")
         call pargi(linenum)
      }
      else {

         switch (model_index) {
         
         case RUMPH_INDEX:

             ismflux = flux * rumphism(wav,hcol,heicol,heiicol)

         default:
             call error(111,"Invalid model selected")
         }

         call fprintf(out_fd,"%10.6f %.6f\n")
         call pargd(wav)
         call pargd(ismflux)
    
         # force immediate printout in case we're interactive
         call flush(out_fd)
      }
   }

  call close(in_fd)
  call close(out_fd)


  # free the stack
  call sfree(stackp)


end

