# Copyright (c) 1991 - 1900.  The Regents of the University of
# California (Regents).  All Rights Reserved.
# H+
# 	Title	:  psf_convolve.x
# 	Author	:  tnk
# 	Date	:  04/28/94
# 	Synopsis:  
# 	Keywords:
# 		@(#) psf_convolve.x -r1.1 04/28/94 UCB SSL CEA EGOCS
# 	Revisions:
# 	mm/dd/yy   name		description
# 	04/28/94   tnk	date and time created 94/04/28 14:48:02 by tnk
# H-
# 
#

include <dfset.h>

# @(#)psf_convolve.x	1.1 4/28/94 UCB CEA EGO

# convolve a signal df with a set of gaussians.
# assumes signal is regularly spaced.
# a different gaussian is used for each element of the signal.
# the gaussians are specified by a DF containing a list of FWHM values
#
# for efficiency, takes the signal df, stuffs it into a zero
# padded work array in memory, and does the convolution outside of the
# df library.
#
# the work array is zero padded to the length of the convolution
# kernal on each end, so that we easily handle convolution end effects
#
#  ---------------------------------------------
# | zero-pad |      signal          | zero-pad |   <-- "work" buffer
#  ---------------------------------------------
# |<-----------worklen ----------------------->|
# |<-kernal->|
#      | <-- convolution --------------->|
#
#
# |<----------------psf fwhm ----------------->|   <-- "psf_y" buffer
#
# result is placed back into the incoming signal df

define FWHM_TO_SIGMA 2.355  # conversion factor

procedure psf_convolve(signal_df,x_spacing,psf_fwhm_list,n_sigma)
pointer signal_df         #I/O DF signal to convolve on (PD)
double  x_spacing         #I spacing of signal domain
pointer psf_fwhm_list     #I DF (PF) list of PSF FWHM values (kernal fwhm's)
double  n_sigma           #I no. of sigma out to create gaussian kernal

pointer psf_match_list
int     siglen   # length of signal array (PD)
int     kernlen  # length of kernal array
int     worklen  # length of work array
pointer kern     # kernal array
pointer work_x   # working array (x)
pointer work_y   # working array (y)
pointer out_y    #
pointer psf_y    # PSF array (y)
pointer sp       # stack pointer
double  xmin, xmax, ymin, ymax
double  max_fwhm
double  width
int     n_sig_steps
int     n_pad_steps
int     n_tot_steps
double  x0
int     nread
pointer new_sig

# FUNCTIONS
pointer df_open()
int     df_len()
int     df_dgetd()

begin

   call smark(sp)

   # determine kernal length by figuring out how many
   # x steps it takes to contain N_SIGMA for the widest kernal
   call df_minmax(psf_fwhm_list,xmin,xmax,ymin,max_fwhm)
   width = 2 * (max_fwhm / FWHM_TO_SIGMA) * n_sigma
   kernlen = width / x_spacing

   new_sig = df_open(DF_PD,TY_DOUBLE,TY_DOUBLE)
   call df_pedge(new_sig,0.0d0,0.0d0)

   siglen = df_len(signal_df)
   call df_minmax(signal_df,xmin,xmax,ymin,ymax)

   # change the signal domain so that it's evenly spaced
   # (incidentally, this also pads with zeros on the right & left
   # assuming the df edges are zero)
   n_sig_steps=(xmax-xmin)/x_spacing+1
   n_pad_steps=kernlen
   x0 = xmin - (x_spacing * n_pad_steps)
   n_tot_steps = n_sig_steps + n_pad_steps*2 + 1 
   call df_rsdomain(signal_df,x0,x_spacing,n_tot_steps,new_sig)

   psf_match_list = df_open(DF_PF,TY_DOUBLE,TY_DOUBLE)
   call df_pedge(psf_match_list,0.0d0,0.0d0)

   # make the psf fwhm list match the domain of the signal
   call df_dfdomain(psf_fwhm_list, new_sig, psf_match_list)   

   worklen = n_tot_steps
   call salloc(work_x,worklen,TY_DOUBLE)
   call salloc(work_y,worklen,TY_DOUBLE)
   call salloc(out_y, worklen,TY_DOUBLE)  
   call salloc(psf_y, worklen,TY_DOUBLE)
   call salloc(kern,  kernlen,TY_DOUBLE)

   # zero buffers
   call aclrd(Memd[work_x],worklen)
   call aclrd(Memd[work_y],worklen)
   call aclrd(Memd[out_y], worklen)
   call aclrd(Memd[psf_y], worklen)
   call aclrd(Memd[kern],  kernlen)

   # convert DFs to arrays
   nread = df_dgetd(new_sig,Memd[work_x],Memd[work_y],worklen)
   nread = df_dgetd(psf_match_list,Memd[work_x],Memd[psf_y],worklen)

   call arr_conv(Memd[work_x],Memd[work_y],worklen,
     Memd[out_y],Memd[psf_y],Memd[kern],kernlen)

   # transfer convolution back into signal DF 
   call df_reset(signal_df)
   call df_dputd(signal_df,Memd[work_x],Memd[out_y],worklen)
   call df_pedge(signal_df,0.0d0,0.0d0)
   call df_px0(signal_df,Memd[work_x]-x_spacing)

   call df_close(psf_match_list)
   call df_close(new_sig)

   call sfree(sp)

end

# do a changing kernal convolution on data in arrays
procedure arr_conv(work_x,work_y,worklen,out_y,psf_y,kern,kernlen)
double  work_x[ARB]  #I "work buffer" x values
double  work_y[ARB]  #I work buffer y values (contains signal & padding)
int     worklen      #I length of work buffer
double  out_y[ARB]   #O output buffer (length=worklen)
double  psf_y[ARB]   #I PSF FWHM values 
double  kern[ARB]    #I kernal buffer
int     kernlen      #I length of kern buffer

# assumes domain of psf_match_list already matches domain of work_x

int     start
int     half_kernlen
double  y_convolve
double  xmid     # x midpt of gaussian
double  sigma    # one sigma value of gaussian
int     i
int     k
int     j
double  old_x
double  new_y

begin


   # for each bin
   #   generate a gaussian of the associated fwhm centered at the bin
   #   convolve the gaussian with the signal (ie, multiply & integrate)
   #   save off the new point

   # note that we start convolving halfway into the kernal-length padding

   # convolve 
   half_kernlen = kernlen/2
   start = half_kernlen
   for (i=start; i<worklen-half_kernlen; i=i+1) {

      # generate a guassian kernal
      sigma = psf_y[i] / FWHM_TO_SIGMA
      xmid = work_x[i]
      call mkgauss(xmid,sigma,kernlen,work_x[i-half_kernlen],kern)

      # multiply signal by kernal and integrate
      k = 1
      j = i+k-1-half_kernlen
      old_x = work_x[j]
      y_convolve = 0.0d0
      for (k=2; k<=kernlen; k=k+1) {
         j = i+k-1-half_kernlen
         new_y = kern[k] * work_y[j]
         y_convolve = y_convolve + new_y * (work_x[j] - old_x)
         old_x = work_x[j]
      }
      out_y[i] = y_convolve
   }

end
