# Copyright (c) 1991 - 1900.  The Regents of the University of
# California (Regents).  All Rights Reserved.
# H+
# 	Title	:  calc_cols.x
# 	Author	:  tnk
# 	Date	:  12/10/92
# 	Synopsis:  
# 	Keywords:
# 		@(#) calc_cols.x -r1.6 12/10/92 UCB SSL CEA EGOCS
# 	Revisions:
# 	mm/dd/yy   name		description
# 	12/21/99   bboyd	diagnostics added and then removed
# 	01/13/97   tnk	took out booleans controlling output columns
# 	08/09/95   tnk	changed from DSS to "Look" (made more generic)
# 	09/20/94   tnk	adjusted error handling (indef time)
# 	12/17/92   tnk	added ram angle & velocity vectors
# 	12/10/92   tnk	date and time created 92/12/10 11:55:06 by tnk
# H-
# 
#


# T. Kilsdonk
# 8 Dec 1992
# @(#)calc_cols.x	1.6 12/21/99

# INCLUDE FILES
include <tbset.h>
include <math.h>

# define the column information
include "backmon.h"

# calculate the values needed for various background monitors
# input table contains aspect & position information
# output table gets all the new monitors

procedure calc_cols(in_tab,skip,asp_col,time_col,pos_col,out_tab,
   out_col, bore_qfor, max_time_diff)
pointer in_tab       #I input table pointer
int     skip         #I input skip factor
pointer asp_col[4]   #I input aspect W-Z columns
pointer time_col     #I time column
pointer pos_col[3]   #I position x,y,z columns
pointer out_tab      #O output table pointer
pointer out_col[ARB] #O output column pointers
double  bore_qfor[4] #I boresight forward quaternion
double  max_time_diff #I max time difference (secs) for vel-view & s/c veloc.

# *** local vars ***
int num_rows        # No. of rows on the input table
int in_row          # input row number
int out_row         # output row number
double time         # aspect time
double time_old     # last time
double time_diff    # diff in time
bool time_bad       # flag to check for undefined element
bool asp_bad[4]     # flag to check for undefined elements
bool pos_bad[3]     # flag to check for undefined elements
double asp[4]       # w,x,y,z of aspect
double pos[3]       # position value (in meters) x,y,z
double pos_old[3]   # old position vector
double pos_diff[3]  # diff in position vectors
double velocity[3]  # satellite velocity vector
double height       # height of satellite (in km)
double mjd          # modified julian date
double ra           # right ascension
double dec          # declination
double gmpos        # greenwich meridan position (radians)
double gmpos_deg    # position of GM in degrees
double sun_long     # solar geographic longitude
double sun_lat      # solar geographic latitude
double sc_long      # spacecraft long
double sc_lat       # spacecraft lat
double look_long     # dss look dir long
double look_lat      # dss look dir lat
double ramdir_long  # ram direction longitude
double ramdir_lat   # ramdirection latitude
double vel_view_ang # velocity view angle (view direction to ram direction)
double asun_zen_ang # antisolar_zenith_angle  (from S/C zenith to antisun)
double look_zen_ang  # dss_zenith_angle
double look_asun_ang # dss_antisun_angle
double mlat         # magnetic latitude
double mmjd         # mod julian date - 40000

# FUNCTIONS
int    tbpsta()     # get value of a parameter
double greenra()    # RA of Greenwich meridian at given mjd
double arcdist()    # arc angle between two points

# reads in time, aspect quaternion, and position x,y,z
# puts out time & satellite height etc.

# Aspect time is in seconds from (MJD + 40000)
# where MJD is modified julian Date,
# MJD epoch = 16 Nov 1858 Noon (JD=2,400,000.5)

define MJD_OFFSET           40000.0D0    # offset in telemetry from MJD
define SECS_PER_DAY         86400.0D0    
define SECS_PER_HOUR         3600.0D0
define POS_VECLEN               3        # length of position/veloc vectors

begin


  # find num of rows in input table
  num_rows = tbpsta(in_tab, TBL_NROWS )
  call aclrd(pos,POS_VECLEN)         # zero position array
  time = 0.0D0

  # go down the input rows
  out_row = 0
  do in_row = 1, num_rows, skip {

     # save off old position & time
     call amovd(pos,pos_old,POS_VECLEN)
     time_old = time                          

     # read time, aspect (w,x,y,z), position (x,y,z) from input table
     call tbrgtd(in_tab, time_col, time, time_bad, 1, in_row) 
     call tbrgtd(in_tab,  asp_col,  asp,  asp_bad, 4, in_row) 
     call tbrgtd(in_tab,  pos_col,  pos,  pos_bad, 3, in_row) 

     if (time_bad   || (time == INDEFD) || 
         asp_bad[1] || asp_bad[2] || asp_bad[3] || asp_bad[4] ||
         pos_bad[1] || pos_bad[2] || pos_bad[3])
     {
        # NOOP
     }
     else {

        out_row = out_row + 1

        # <<< Time >>>
        call tbrptd(out_tab, out_col[C_TIME], time, 1, out_row)
        # <<< MMJD >>>
        # convert aspect time in seconds to Modified Julian Date
        mjd =  time / SECS_PER_DAY  + MJD_OFFSET
        mmjd = mjd - MJD_OFFSET    # modified-modified julian date 
        call tbrptd(out_tab, out_col[C_MMJD], mmjd, 1, out_row)  

        # Get position of Greenwich meridian at specified
        # modified Julian Date.
        # Needed for ephemeris later.
        gmpos = greenra(mjd)

        # <<< LOOKLong, LOOKLat >>>
        # convert aspect to LOOK boresight ra/dec in radians
        call bore_radec(asp,bore_qfor,ra,dec)

        #convert ra/dec in radians to long/lat in degrees 
        call radec2ll(ra,dec,look_long,look_lat,gmpos)
        call tbrptd(out_tab, out_col[C_LOOKLONG], look_long, 1, out_row) 
        call tbrptd(out_tab, out_col[C_LOOKLAT],  look_lat,  1, out_row) 

        # <<< SatLong, SatLat, SatHt >>>

        #convert satellite position (ephemeris x,y,z) to long/lat
        call carsph(pos,ra,dec)                      # xyz to ra/dec
        call radec2ll(ra,dec,sc_long,sc_lat,gmpos)   # ra/dec to long/lat deg
        call tbrptd(out_tab, out_col[C_SATLONG], sc_long, 1, out_row) 
        call tbrptd(out_tab, out_col[C_SATLAT],  sc_lat,  1, out_row) 

        # height of satellite is just magnitude of position 3-vector
        # put out in km
        height = SQRT(pos(1)*pos(1)+pos(2)*pos(2)+pos(3)*pos(3)) / 1000.0D0
        call tbrptd(out_tab, out_col[C_SATHT], height, 1, out_row) 

        # <<< SunLong, SunLat >>>
        call sunpos(mjd,ra,dec)

        #convert ra/dec to long/lat in degrees
        call radec2ll(ra,dec,sun_long,sun_lat,gmpos)
        call tbrptd(out_tab, out_col[C_SUNLONG], sun_long, 1, out_row) 
        call tbrptd(out_tab, out_col[C_SUNLAT],  sun_lat,  1, out_row) 

        # <<< ASunZen >>>
        # find antisolar zenith angle
        # (angle from antisun-position to spacecraft position)
        asun_zen_ang = 180.0d0 - arcdist(sun_long,sun_lat,sc_long,sc_lat)
        call tbrptd(out_tab, out_col[C_ASUNZENITH], asun_zen_ang, 1, out_row) 

        # <<< LOOKZen >>>
        # find look to s/c zenith angle
        # (angle from look look dir to spacecraft zenith)
        look_zen_ang = arcdist(look_long,look_lat,sc_long,sc_lat)
        call tbrptd(out_tab, out_col[C_LOOKZENITH], look_zen_ang, 1, out_row) 

        # <<< LOOKASun >>>
        # find antisolar look look angle
        # (angle from look look-position to antisun direction)
        look_asun_ang = 180.0d0 - arcdist(sun_long,sun_lat,look_long,look_lat)
        call tbrptd(out_tab, out_col[C_LOOKASUN], look_asun_ang, 1, out_row) 

        # Velocity view angle & Spacecraft velocity vector
        # angle between telescope look dir & the satellite motion (ram) dir

        # put out INDEFs for values we can't compute
        vel_view_ang = INDEFD
        call amovkd(INDEFD,velocity,POS_VECLEN)

        time_diff = time - time_old    # time in seconds
        # only compute if time gap is small (e.g., within 1 orbit)
        if (time_diff < MAX_TIME_DIFF) {
           # << LOOK velocity view angle >>
           call asubd(pos,pos_old,pos_diff,POS_VECLEN)   # position change
           call carsph(pos_diff,ra,dec)                  # xyz to ra/dec
           call radec2ll(ra,dec,ramdir_long,ramdir_lat,gmpos)  # to long/lat
           vel_view_ang = arcdist(look_long,look_lat,ramdir_long,ramdir_lat)

           # << Spacecraft velocity vector >>
           # divide change in position by change in time to get m/s

           call adivkd(pos_diff,time_diff,velocity,POS_VECLEN)
           # change to m/hr
           call amulkd(velocity,SECS_PER_HOUR,velocity,POS_VECLEN)   
           # change to km/hr
           call adivkd(velocity,1000.0D0,velocity,POS_VECLEN)   
        }

        call tbrptd(out_tab, out_col[C_LOOKVELVIEW],vel_view_ang,1, out_row) 

        call tbrptd(out_tab, out_col[C_SATVELX],velocity[1],1, out_row) 
        call tbrptd(out_tab, out_col[C_SATVELY],velocity[2],1, out_row) 
        call tbrptd(out_tab, out_col[C_SATVELZ],velocity[3],1, out_row) 


        # <<< GMPOS >>>
        # convert from radians to degrees
        gmpos_deg = RADTODEG(gmpos)
        call tbrptd(out_tab, out_col[C_GMPOS],  gmpos_deg,  1, out_row) 
       
        # <<< MAGLAT >>>
        # find magnetic latitude (deg) corresponding to the S/C Geo long/lat
        call maglat(sc_long,sc_lat,mlat)
        call tbrptd(out_tab, out_col[C_MAGLAT],  mlat,  1, out_row) 

     }
  }
  

end

