# Copyright (c) 1991 - 1900.  The Regents of the University of
# California (Regents).  All Rights Reserved.
# H+
# 	Title	:  rlintegrate.x
# 	Author	:  tnk
# 	Date	:  08/15/95
# 	Synopsis:  
# 	Keywords:
# 		@(#) rlintegrate.x -r1.4 08/15/95 UCB SSL CEA EGOCS
# 	Revisions:
# 	mm/dd/yy   name		description
# 	10/10/95   mabbott	added check for nearby data when in a gap, to lessen invalid time
# 	08/15/95   tnk	renamed rlintegrate to avoid collision
# 	08/15/95   tnk	added sccs keywords
# 	08/15/95   tnk	date and time created 95/08/15 14:43:50 by tnk
# H-
# 
#

#  @(#)rlintegrate.x	1.4 10/10/95 UCB CEA
#
#  This is a subroutine to perform effective exposure time computation.
#
#  The caller is responsible for providing the time bounds to integrate over;
#  a run-list encoded set of time intervals which are the "valid" times for
#  the observation; arrays of primbsch correction times and values; arrays
#  of detector position times and values; an array with a vignetting correction
#  image and the associated MWCS info; an array with the distortion correction
#  image and the associated MWCS info; a pointer to a pixel mask; a detector
#  quadrant boundary info structure; and various flags and output variables.
#
#  The task effexp uses this routine; see the code in t_effexp.x for examples
#  of how to read all of the required input tables and images.
#
#  This source file depends on  - the binary search routine in the
#  file hunt.f; the polynomial interpolation code in polint.f, both of those
#  are derived from Numerical Recipes.  It also depends on the quadrant
#  boundary code in qbound.x
#
#
# Makes use of Numerical Recipes library routines,
# Copyright (C) 1986 Numerical Recipes Software
#

include <imhdr.h>

define EPS double(1.0e-5)		# comparison epsilon for floats
define PINTLEN double(2.048)	# reporting interval of primbsch corrections
define VINTLEN double(1.024)	# reporting interval of detector positions

#
#	Integrate a run length encoded segment with corrections
#
procedure	rl_integ(tmin, tmax, xs, xe, xlen, ptime, pcorr, np, 
						 quad2use, dtime, dx, dy, 
						 nd, vig, nvx, nvy, vct, dist, ndx, ndy, dct, 
						 detptr, pim, usevig, undistort, sublen, effexp, 
						 rawexp, badexp)

double	tmin, tmax			#I minimum and maximum time to integrate over

double	xs[xlen], xe[xlen]	#I run-length encoded valid times; start and end
int		xlen				#I length of start and end arrays

double	ptime[np]			#I times of primbsch correction values
real	pcorr[np,4]			#I primbsch correction values
int		np					#I number of primbsch correction values
int		quad2use			#I quadrant to use instead of detpos position

double	dtime[nd]			#I times of detector position values
real	dx[nd]				#I detector x position values
real	dy[nd]				#I detector y position values
int		nd					#I number of detector position values

real	vig[nvx,nvy]		#I vignetting image
int		nvx,nvy				#I size of vignetting image
pointer vct					#I MWCS transformation pointer for vignetting img

real	dist[ndx,ndy]		#I distortion image
int		ndx,ndy				#I size of distortion image
pointer dct					#I MWCS transformation pointer for distortion img

pointer	detptr				#I detector quad structure

pointer	pim					#I pixel mask image pointer

bool	usevig				#I use vignetting map?
bool	undistort			#I undistort before vignetting?

double	sublen				#I length of integration subintervals

double	effexp				#O effective exposure
double	rawexp				#O raw exposure
double	badexp				#O invalid exposure (no corrections)

int 	i, j, k, s, nsub
double	time, endsub, t0, t1
real	cv
bool	invalid

real	corr()
bool	eq_epsilon()

begin
  # Exposure undefined due to open range or no time filter
  if (IS_INDEFD(xs[1]) || IS_INDEFD(xe[xlen]) || xlen<1) {
	effexp = INDEFD
	return
  }

  i = 1   # index into QPOE time intervals
  j = 0   # index into primbsch corrections
  k = 0   # index into detector positions
  s = 1   # subinterval index
  effexp = 0.
  rawexp = 0.
  badexp = 0.

  while (i <= xlen) {
	# Consider only the time between tmin and tmax
	if (xe[i] <= tmin || xs[i] >= tmax) {
	  i = i + 1
	  next
	}

	if (xs[i] < tmin) t0 = tmin
	else t0 = xs[i]

	if (xe[i] > tmax) t1 = tmax
	else t1 = xe[i]

	# Update the raw exposure
	rawexp = rawexp + (t1-t0)

	# Sum subintervals
	nsub = int((t1 - t0) / sublen)
	time = t0 - sublen/2.0
	do s = 1, nsub {
	  time = time + sublen

	  cv = corr(ptime, pcorr, np, quad2use, dtime, dx, dy, nd, time, 
				vig, nvx, nvy, vct, 
				dist, ndx, ndy, dct, detptr, pim, usevig, undistort, k, j, 
				invalid)
	  effexp = effexp + sublen * cv
	  if (invalid) {
		badexp = badexp + sublen
	  }
	}

	# Handle possible partial subinterval at end of interval
	endsub = time+sublen/2.0
	if (! eq_epsilon(endsub, t1, EPS)) {
	  time = endsub + (t1 - endsub) / 2.0
	  cv = corr(ptime, pcorr, np, quad2use, dtime, dx, dy, nd, time, 
				vig, nvx, nvy, vct, 
				dist, ndx, ndy, dct, detptr, pim, usevig, undistort, k, j, 
				invalid)
	  effexp = effexp + (t1 - endsub) * cv
	  if (invalid) {
		badexp = badexp + (t1 - endsub)
	  }
	}

	i = i + 1
  }

  return
end



#
# Return the total correction at a particular time
#
real	procedure corr(ptime, pcorr, np, quad2use, dtime, x, y, n, t, 
					   vig, nvx, nvy, 
					   vct, dist, ndx, ndy, dct, detptr, pim, usevig, 
					   undistort, i, j, invalid)

double	ptime[np]		#I times of primbsch correction values
real	pcorr[np,4]		#I primbsch correction values
int		np				#I number of primbsch correction values
int		quad2use		#I quadrant to use instead of detpos position

double	dtime[n]		#I times of detector positions
real	x[n], y[n]		#I detector positions
int		n				#I number of positions

double	t				#I desired time

real	vig[nvx,nvy]	#I vignetting image 
int		nvx,nvy			#I size of vignetting image
pointer	vct				#I vignetting image MWCS transformation pointer

real	dist[ndx,ndy,2]	#I distortion image 
int		ndx,ndy			#I size of distortion image
pointer	dct				#I distortion image MWCS transformation pointer

pointer	detptr			#I detector quad structure

pointer	pim				#I mask image pointer

bool	usevig			#I apply vignetting?
bool	undistort		#I undistort before vignetting?

int		i				#I/O index to detector positions
int		j				#I/O index to primbsch corrections
bool	invalid			#I/O no correction available, returned value of 1.0

int		k, npts, quad, ix, iy
real	dx, dxerr, dy, dyerr, corr, pc, vc, frac, dxp, dyp

int		pb_quadrant()
bool	eq_epsilon(), im_pmsne2()
real	vcorr_lookup(), pval()

define doprimbsch_ 11
define dovignetting_ 12

begin
  # Do we have a detpos table?
  if (n != 0) {
	# The next section of code attempts to make the most accurate
	# estimate it can of the detector position at the time t.  If it cannot
	# find a position for any reason, the time is considered invalid
	# and the total correction is set to 1.0, regardless of the primbsch
	# correction for that time.

	# find t in the detpos table
	call hunt(dtime, n, t, i)

	# we are off the table.  this time is invalid.
	if (i <= 0 || i >= n) {
	  invalid = true
	  return 1.0
	}
					 
	# Uh oh, we cannot do any interpolation if we are sitting directly
	# in a gap in the times in the detpos table.  Extrapolation is not a
	# good idea.  If we are within a reasonable amount of time of a table
	# entry, just assume we are at the same position as that entry, otherwise
	# this is an invalid time.
	if (! eq_epsilon(dtime[i+1]-dtime[i], VINTLEN, EPS)) {
	  if (abs(t-dtime[i]) < 2*VINTLEN) {
		dx = x[i]
		dy = y[i]
		goto dovignetting_
	  } else if (abs(t-dtime[i+1]) < 2*VINTLEN) {
		dx = x[i+1]
		dy = y[i+1]
		goto dovignetting_
	  } else {
		invalid = true
		return 1.0
	  }
	}

	# good idea.  This is an invalid time.
	#if (! eq_epsilon(dtime[i+1]-dtime[i], VINTLEN, EPS)) {
    #  invalid = true
	#  return 1.0
	#}

	# There is an INDEF detector position before or after this time.  This
	# means we are way off the detector.  The vignetting correction is
	# zero or the mask value is zero, so return zero if either is being used.
	# If neither is being used, then set the vignetting correction to 1.0 and
	# skip to the primbsch correction computation.
	if (IS_INDEFR(x[i]) || IS_INDEFR(x[i+1]) || IS_INDEFR(y[i]) || IS_INDEFR(y[i+1])) {
	  if (usevig || pim != NULL) {
		invalid = false
	    return 0.0
	  }
	  else {
	    vc = 1.0
		goto doprimbsch_
	  }
	}

    # Now we try to do interpolation over 4 pts, taking care to avoid gaps 
	# or INDEFs in between the points.  Simple linear interpolation is not 
	# good enough because the positions can change quite rapidly with time
	# in survey mode, in a non-linear (but smooth) manner.
	if (i<2 || i>n-2 || ! eq_epsilon(dtime[i+2]-dtime[i-1], 3.0*VINTLEN, EPS) ||
		IS_INDEFR(x[i-1]) || IS_INDEFR(x[i+2]) || IS_INDEFR(y[i-1]) || 
		IS_INDEFR(y[i+2])) {

	  # We have a gap or INDEF during the nominal interpolation interval
	  if (i<2 || ! eq_epsilon(dtime[i]-dtime[i-1], VINTLEN, EPS) ||
		  IS_INDEFR(x[i-1]) || IS_INDEFR(y[i-1])) {

		# Point i-1 is unusable

		if (i>n-2 || ! eq_epsilon(dtime[i+2]-dtime[i+1], VINTLEN, EPS) ||
			IS_INDEFR(x[i+2]) || IS_INDEFR(y[i+2])) {

		  # Point i+2 is bad also

		  # Case 3: gaps or INDEFs affecting both pts i-1 and i+2.  
		  # Do linear interpolation from i to i+1
		  frac = (t-dtime[i])/(dtime[i+1]-dtime[i])
		  dx = x[i] + frac * (x[i+1]-x[i])
		  dxerr = 0.0
		  dy = y[i] + frac * (y[i+1]-y[i])
		  dyerr = 0.0
		  k = 0
		  npts = 0
		}
		else {

		  # Case 1: a gap between pts i-1 & i or i-1 is INDEF
		  # Do 3 pt interpolation from i to i+2
		  k = i
		  npts = 3
		  call polint(dtime[k], x[k], npts, t, dx, dxerr)
		  call polint(dtime[k], y[k], npts, t, dy, dyerr)
		}
	  }
	  else {
		# Must be a problem with point i+2

		# Case 2: a gap between pts i+2 & i+1 or i+2 is INDEF
		# Do 3 pt interpolation from i-1 to i+1
		k = i-1
		npts = 3
		call polint(dtime[k], x[k], npts, t, dx, dxerr)
		call polint(dtime[k], y[k], npts, t, dy, dyerr)
	  }
	}
	else {
	  # Whew, made it!  We have 4 valid points, so do 4 pt interpolation.
	  k = i-1
	  npts = 4
	  call polint(dtime[k], x[k], npts, t, dx, dxerr)
	  call polint(dtime[k], y[k], npts, t, dy, dyerr)
	}

dovignetting_

    # At this point we have a valid detector position.

	# Check mask, if any.  If not in mask we are done.  Detector coords
	# run from 0-2047, while mask pixels are from 1-2048.  Note that we
    # are assuming that the user has specified the mask and detpos table
    # consistently in either distorted or undistorted coordinates.  We assume
    # all points outside the mask have a mask value of zero and are rejected.
	if (pim != NULL) {
	  ix = nint(dx)+1
	  iy = nint(dy)+1

	  if (ix<1 || ix>IM_LEN(pim,1) || iy<1 || iy>IM_LEN(pim,2)) {
		invalid = false
		return 0.0
	  }
	  if (! im_pmsne2(pim, ix, ix, iy, iy)) {
		invalid = false
		return 0.0
	  }
	}


	# If usevig is false, this is a mask only run (or the user has just wasted
    # our time)
	if (usevig) {
      if (undistort) {
		call det_undistort(dist, ndx, ndy, dct, dx, dy, dxp, dyp)
	  }
	  else {
		dxp = dx
		dyp = dy
	  }

	  # look up the vignetting correction.  if either dx or dy are INDEF
	  # from detpos table, this will return 0.0
	  vc = vcorr_lookup(vig, nvx, nvy, vct, dxp, dyp, pim)
	}
	else vc = 1.0

  }
  else {
	# No detector positions available, cannot check mask or vignetting
	vc = 1.0
  }

doprimbsch_

  # Get the primbsch correction value, if present.  Use the given fixed 
  # quadrant, if present
  if (np != 0) {
	if (quad2use >= 0) {
	  quad = quad2use
	} else {
	  quad = pb_quadrant (detptr, dx, dy)
	}
    pc = 1.0/pval(ptime, pcorr[1,quad+1], np, t, j, invalid)
  }
  else {
	invalid = false
	pc = 1.0
  }
  corr = pc * vc

  return corr
end



#
# Return the primbsch/deadtime correction at a particular time
#
real	procedure pval(time, corr, n, t, i, invalid)

double	time[n]			#I times of correction values
real	corr[n]			#I correction values
int		n				#I number of correction values
double	t				#I desired time
int		i				#I/O index
bool	invalid			#I/O no correction available

bool	lt_epsilon()

begin
  invalid = true

  # If we have no primbsch table, stop here.
  if (n == 0) {
	invalid = false
	return 1.0
  }

  # Find t in the primbsch table
  call hunt(time, n, t, i)

  # The time is marked as invalid if we are outside the table, or in a gap
  # in the table, or if the correction value is INDEF.
  if (i >= n) return 1.0
  if (lt_epsilon(t, time[i+1]-PINTLEN, EPS)) return 1.0
  if (IS_INDEFR(corr[i+1])) return 1.0

  invalid = false

  # Catch any 0 corrections.  These should be impossible but occur due to the
  # famous negative primbsch problem.
  if (corr[i+1] == 0.0) return 1.0

  # The correction value applies to the previous PINTLEN seconds.  We do not
  # need to interpolate.
  return corr[i+1]
end



#
#  Return the vignetting correction at a detector position
#
real	procedure vcorr_lookup(vig, nvx, nvy, ct, x, y)

real	vig[nvx,nvy]	#I vignetting data
int		nvx,nvy			#I size of vignetting array
pointer ct				#I image MWCS transformation pointer
real	x,y				#I position to look up

int		ix, iy
real	y1, y2, y3, y4, t, u
real	lx, ly, corr

begin
  if (IS_INDEFD(x) || IS_INDEFD(y)) {
	return 0.0
  }

  # Convert from detector position to logical coords in vignetting image
  call mw_c2tranr(ct, x, y, lx, ly)

  ix = int(lx)
  iy = int(ly)

  if (ix < 1 || iy < 1 || ix >= nvx || iy >=nvy) return 0.0

  t = lx - ix
  u = ly - iy

  y1 = vig[ix,iy]
  y2 = vig[ix+1,iy]
  y3 = vig[ix+1,iy+1]
  y4 = vig[ix,iy+1]

  # Bilinear interpolation

  corr = (1.0-t)*(1.0-u)*y1 + t*(1.0-u)*y2 + t*u*y3 + (1.0-t)*u*y4

  return corr
end


#
# Apply distortion correction to a pixel
#
procedure det_undistort(dist, nx, ny, ct, dx, dy, dxp, dyp)

real	dist[nx,ny,2]	#I distortion image 
int		nx,ny			#I size of distortion image
pointer	ct				#I distortion image MWCS transformation pointer
real	dx, dy			#I detector coordinates
real	dxp, dyp		#O undistorted coordinates

int		ix, iy
real	y1, y2, y3, y4, t, u, delta
real	lx, ly

begin
  if (IS_INDEFD(dx) || IS_INDEFD(dy)) {
	dxp = INDEFD
	dyp = INDEFD
	return
  }

  # Convert from detector position to logical coords in distortion image
  call mw_c2tranr(ct, dx, dy, lx, ly)

  ix = int(lx)
  iy = int(ly)

  if (ix < 1 || iy < 1 || ix >= nx || iy >=ny) {
	dxp = dx
	dyp = dy
	return
  }

  t = lx - ix
  u = ly - iy

  # Bilinear interpolation

  y1 = dist[ix,iy,1]
  y2 = dist[ix+1,iy,1]
  y3 = dist[ix+1,iy+1,1]
  y4 = dist[ix,iy+1,1]

  # The distortion images contain the delta from distorted to undistorted
  # coordinates.
  delta = (1.0-t)*(1.0-u)*y1 + t*(1.0-u)*y2 + t*u*y3 + (1.0-t)*u*y4
  dxp = dx + delta

  y1 = dist[ix,iy,2]
  y2 = dist[ix+1,iy,2]
  y3 = dist[ix+1,iy+1,2]
  y4 = dist[ix,iy+1,2]

  # The distortion images contain the delta from distorted to undistorted
  # coordinates.
  delta = (1.0-t)*(1.0-u)*y1 + t*(1.0-u)*y2 + t*u*y3 + (1.0-t)*u*y4
  dyp = dy + delta
end
