# Copyright (c) 1991 - 1900.  The Regents of the University of
# California (Regents).  All Rights Reserved.
# H+
# 	Title	:  t_parallax.x
# 	Author	:  tnk
# 	Date	:  11/11/96
# 	Synopsis:  
# 	Keywords:
# 		@(#) t_parallax.x -r1.7 11/11/96 UCB SSL CEA EGOCS
# 	Revisions:
# 	mm/dd/yy   name		description
# 	02/22/97   tnk	added history card containing input filename
# 	01/09/97   tnk	added INDEF time handling for ephemeris & position tables
# 	01/03/97   mabbott	reversed algorithm to interpolate ephem not position
# 	12/27/96   tnk	input ra units now hours or degrees, output same as input
# 	11/15/96   tnk	copies extra columns, minor bug fixes
# 	11/11/96   tnk	made timedel a parameter
# 	11/11/96   tnk	date and time created 96/11/11 10:38:57 by tnk
# H-
# 
#

# @(#)t_parallax.x	1.7 2/22/97 UCB CEA
#
# Compute parallax corrections for nearby source ephemerides
#
#   This algorithm takes as input the position info from table1 of
#   the EUVE data and a source ephemeris.  The ephemeris must
#   contain the ra, dec, and distance of the source vs time during
#   the observation. The ra and dec should be in decimal degrees
#   and the distance from earth center in AU.  The times in the
#   ephemeris are assumed to be corrected for light-travel and must
#   represent the time of the positions as seen at the earth.  The
#   light-travel time from the center of the earth to EUVE is
#   ignored as the error is smaller than the accuracy of the
#   position data.
#

include <tbset.h>
include <fset.h>

define MAX_EXTRA 64

define XPOS 1
define YPOS 2
define ZPOS 3

define RA 1
define DEC 2
define DIST 3

define OTIME 1
define ORA 2
define ODEC 3
define ODIST 4

define EPS double(1.0e-5)

define PI double(3.14159265358979323846264339)
define DEG2RAD double(PI/180.0)
define MPERAU double(149597870000.0)

procedure t_parallax()

int		i, j
int		ephem_row, ephem_cols, euve_row
int	out_row
double	time
double	euvepos[3]
double	ra, dec, dist
double	x, y, z
double	r, h, b, l, p
double	ex, ey, ez
double	out[4]
bool	null[4]
bool	valid
bool    ra_hours
pointer	sp, str
pointer ephem_tp, euve_tp, out_tp
pointer ephem_time, ephe_cp[3]
pointer euve_time, euve_cp[3]
pointer out_cp[4]
pointer	iextras, oextras, xcp, exbuf
pointer units
pointer infile
pointer histbuff

int		tbpsta()
bool	fp_equald()
pointer	tbtopn(), tbcnum()

begin
  call smark(sp)

  call salloc(str, SZ_LINE, TY_CHAR)
  call salloc(infile, SZ_LINE, TY_CHAR)

  call fseti(STDOUT, F_FLUSHNL, YES)

  # open ephemeris file
  call clgstr("ephemeris", Memc[infile], SZ_LINE)
  ephem_tp = tbtopn(Memc[infile], READ_ONLY, 0)
  ephem_row = tbpsta(ephem_tp, TBL_NROWS)

  call clgstr("ephtime", Memc[str], SZ_LINE)
  call tbcfnd(ephem_tp, Memc[str], ephem_time, 1)
  if (ephem_time == 0) call error(123, "ephemeris time column not found")

  call clgstr("racol", Memc[str], SZ_LINE)
  call tbcfnd(ephem_tp, Memc[str], ephe_cp[RA], 1)
  if (ephe_cp[RA] == 0) call error(123, "RA column not found")

  # get input RA units
  call salloc(units, SZ_LINE, TY_CHAR)
  call tbcigt(ephe_cp[RA], TBL_COL_UNITS,  Memc[units],  SZ_LINE)
  if(Memc[units] == 'h' || Memc[units] == 'H')
	ra_hours = true
  else if(Memc[units] == 'd' || Memc[units] == 'D')
	ra_hours = false
  else
  {
     call eprintf("Invalid RA units in input file: %s\n")
     call pargstr(Memc[units])
     call error(1,"Ra units must be hours (H/h) or Degrees (D/d)")
  }

  call clgstr("deccol", Memc[str], SZ_LINE)
  call tbcfnd(ephem_tp, Memc[str], ephe_cp[DEC], 1)
  if (ephe_cp[DEC] == 0) call error(123, "Dec column not found")

  call clgstr("distcol", Memc[str], SZ_LINE)
  call tbcfnd(ephem_tp, Memc[str], ephe_cp[DIST], 1)
  if (ephe_cp[DIST] == 0) call error(123, "Distance column not found")

  # get pointers to all additional columns
  ephem_cols = tbpsta(ephem_tp, TBL_NCOLS)
  if (ephem_cols-4 > MAX_EXTRA) {
	call eprintf("Maximum number of extra columns exceeded.")
	call eprintf("Only the first %d columns will be copied to output.")
	call pargi(MAX_EXTRA)
	ephem_cols = MAX_EXTRA+4
  }
  call salloc(iextras, ephem_cols-4, TY_POINTER)
  j = 0
  do i = 1, ephem_cols {
	xcp = tbcnum(ephem_tp, i)
	if (xcp != ephem_time && xcp != ephe_cp[RA] &&
		xcp != ephe_cp[DEC] && xcp != ephe_cp[DIST]) {
	  Memi[iextras+j] = xcp
	  j = j + 1
	}
	if (j == MAX_EXTRA) break
  }

  # get EUVE position info
  call clgstr("position", Memc[str], SZ_LINE)
  euve_tp = tbtopn(Memc[str], READ_ONLY, 0)
  euve_row = tbpsta(euve_tp, TBL_NROWS)

  call clgstr("postime", Memc[str], SZ_LINE)
  call tbcfnd(euve_tp, Memc[str], euve_time, 1)
  if (euve_time == 0) call error(123, "position time column not found")

  call clgstr("xcol", Memc[str], SZ_LINE)
  call tbcfnd(euve_tp, Memc[str], euve_cp[XPOS], 1)
  if (euve_cp[XPOS] == 0) call error(123, "x position column not found")

  call clgstr("ycol", Memc[str], SZ_LINE)
  call tbcfnd(euve_tp, Memc[str], euve_cp[YPOS], 1)
  if (euve_cp[YPOS] == 0) call error(123, "y position column not found")

  call clgstr("zcol", Memc[str], SZ_LINE)
  call tbcfnd(euve_tp, Memc[str], euve_cp[ZPOS], 1)
  if (euve_cp[ZPOS] == 0) call error(123, "z position column not found")

  # open output ephemeris table
  call clgstr("output", Memc[str], SZ_LINE)
  out_tp = tbtopn(Memc[str], NEW_COPY, ephem_tp)
  call tbtcre(out_tp)

  # add history card containing input table name
  call salloc(histbuff, SZ_LINE, TY_CHAR)
  call sprintf(Memc[histbuff], SZ_LINE, "parallax: input=%s")
  call pargstr(Memc[infile])
  call tbhadt(out_tp, "HISTORY", Memc[histbuff])


  call clgstr("ephtime", Memc[str], SZ_LINE)
  call tbcfnd(out_tp, Memc[str], out_cp[OTIME], 1)
  if (out_cp[OTIME] == 0) call error(123, "ephemeris time column not found")

  call clgstr("racol", Memc[str], SZ_LINE)
  call tbcfnd(out_tp, Memc[str], out_cp[ORA], 1)
  if (out_cp[ORA] == 0) call error(123, "RA column not found")

  call clgstr("deccol", Memc[str], SZ_LINE)
  call tbcfnd(out_tp, Memc[str], out_cp[ODEC], 1)
  if (out_cp[ODEC] == 0) call error(123, "Dec column not found")

  call clgstr("distcol", Memc[str], SZ_LINE)
  call tbcfnd(out_tp, Memc[str], out_cp[ODIST], 1)
  if (out_cp[ODIST] == 0) call error(123, "Distance column not found")

  # get pointers to other cols
  call salloc(oextras, ephem_cols-4, TY_POINTER)
  j = 0
  do i = 1, ephem_cols {
	xcp = tbcnum(out_tp, i)
	if (xcp != out_cp[OTIME] && xcp != out_cp[ORA] &&
		xcp != out_cp[ODEC] && xcp != out_cp[ODIST]) {
	  Memi[oextras+j] = xcp
	  j = j + 1
	}
  }
  call salloc(exbuf, ephem_cols-4, TY_DOUBLE)

  # now process ephemeris
  out_row = 1
  do i = 1, euve_row {

	# read EUVE position (this is stored in the telemetry in
	# J2000 cartesian coordinates in meters)
	call tbrgtd(euve_tp, euve_time, time, null, 1, i)
	call tbrgtd(euve_tp, euve_cp, euvepos, null, 3, i)

	if (IS_INDEFD(time))
	{
	   call eprintf("Warning:  INDEF time in position table, row: %d\n")
	   call pargi(i)
	   next
	}

	# get the source position
	call get_ephem_row(ephem_tp, ephem_time, ephe_cp, Memi[iextras], 
						ephem_cols-4, ephem_row, time, ra, dec, dist, 
						Memd[exbuf], valid)

	if (! valid) {
	  # put out an INDEF ephemeris point
	  l = INDEFD
	  b = INDEFD
	  r = INDEFD
	} else {
	  # coordinate transformation to system with EUVE at origin

	  # if input RA is in hours, convert to degrees
	  if (ra_hours) ra = ra * 15

	  # convert source position in degrees and distance in AU
	  # to cartesian coords in meters
	  x = MPERAU * dist * cos(DEG2RAD*dec) * cos(DEG2RAD*ra)
	  y = MPERAU * dist * cos(DEG2RAD*dec) * sin(DEG2RAD*ra)
	  z = MPERAU * dist * sin(DEG2RAD*dec)

	  # origin change
	  ex = x - euvepos[XPOS]
	  ey = y - euvepos[YPOS]
	  ez = z - euvepos[ZPOS]

	  # convert coordinates back to polar coords in degrees
	  r = sqrt(ex*ex + ey*ey + ez*ez)
	  h = sqrt(ex*ex + ey*ey)
	  if (fp_equald(h, 0.0)) {
		if (fp_equald(ez, 0.0)) b = 0.0
		else if (ez > 0.0) b = 90.0
		else b = -90.0
	  } else {
		b = atan(ez/h) / DEG2RAD
	  }
	  if (fp_equald(ex, 0.0) && fp_equald(ey, 0.0)) {
		l = 0.0;
	  } else {
		p = 2.0 * atan(ey / (abs(ex) + h)) / DEG2RAD
		if (ex < 0.0) {
		  l = 180.0 - p
		} else if (ey < 0.0) {
		  l = 360.0 + p
		} else {
		  l = p
		}	
	  }
	  
	  r = r / MPERAU
	}

	out[ORA] = l
	if (ra_hours && !IS_INDEFD(out[ORA]))
		out[ORA] = out[ORA] / 15
	out[ODEC] = b
	out[ODIST] = r
	out[OTIME] = time

	# write out columns
	call tbrptd(out_tp, out_cp, out, 4, out_row)
	call tbrptd(out_tp, Memi[oextras], Memd[exbuf], ephem_cols-4, out_row)
	out_row = out_row + 1
  }
  
  # Clean up
  call tbtclo(ephem_tp)
  call tbtclo(euve_tp)
  call tbtclo(out_tp)

  call sfree(sp)
end


# return a row from the ephemeris table
procedure get_ephem_row(tp, tcp, ecp, exp, nex, npos,  t, ra, dec, 
						dist, extra, valid)

pointer	tp			# ephemeris table pointer
pointer	tcp			# ephemeris table time column pointer
pointer	ecp[3]		# ephemeris table position column pointers
pointer exp[nex]	# ephemeris table extra column pointers
int		nex			# number of extra columns
int		npos		# ephemeris table length
double	t			# time at which to interpolate
double	ra, dec, dist	# output coordinates
double  extra[nex]  # output extra data
bool	valid		# output valid flag

int		i

int		row
double	time, pos[3], frac, lasttime, lastpos[3]
double  lastextra[MAX_EXTRA], readextra[MAX_EXTRA]
bool	firsttime
bool	null[MAX_EXTRA]

data	firsttime /true/

begin

  # set up when first called
  if (firsttime) {
	firsttime = false

	row = 0

	lasttime = 0.0
	time = 0.0
	do i = 1, 3 {
	  lastpos[i] = 0.0
	  pos[i] = INDEFD
	}
	do i = 1, nex {
	  lastextra[i] = 0.0
	  readextra[i] = INDEFD
	}
  }


  # first get the bounding rows in position table

  # are we still within the previous bounds?
  if (t > time) {
	# our previous bounds are no longer useful, get new ones
	if (row < npos) {
	  # advance to table rows spanning desired time
	  repeat {
		lasttime = time
		do i = 1, 3 {
		  lastpos[i] = pos[i]
		}
		do i = 1, nex {
		  lastextra[i] = readextra[i]
		}
		row = row + 1
		call tbrgtd(tp, tcp, time, null, 1, row)
		call tbrgtd(tp, ecp, pos, null, 3, row)
		call tbrgtd(tp, exp, readextra, null, nex, row)

	    	if (IS_INDEFD(time))
		{
		   call eprintf("Warning:  INDEF time in ephemeris table, row: %d\n")
		   call pargi(row)
		   time = 0.0
		}

	  } until (time > t || row == npos)
	
	  # Check if we are off the table and the time is invalid.
	  if (row == 1) {
		# the desired time was before the first table row
		# start over the next time
		row = 0
		valid = false
		return
	  }
	  if (row == npos && time <= t) {
		# the desired time was after the last table row
		valid = false
		return
	  }
	} else {
	  # we are already at end of table
	  if (time <= t) {
		# the desired time was after the last table row
		valid = false
		return
	  }
	}
  }

  # now check the validity of the rows we have reached

  do i = 1, 3 {
	if (IS_INDEFD(lastpos[i]) || IS_INDEFD(pos[i])) {
	  # There is an INDEF position before or after this time.  This means
	  # the ephemeris information is missing.  The time is invalid.
	  valid = false
	  return
	}
  }
  do i = 1, nex {
	if (IS_INDEFD(lastextra[i]) || IS_INDEFD(readextra[i])) {
	  # There is an INDEF extra column before or after this time.  This means
	  # the ephemeris information is missing.  The time is invalid.
	  valid = false
	  return
	}
  }

  # Interpolate

  valid = true
  frac = (t - lasttime) / (time - lasttime)
  ra = (lastpos[RA] + frac * (pos[RA] - lastpos[RA]))
  dec = (lastpos[DEC] + frac * (pos[DEC] - lastpos[DEC]))
  dist = (lastpos[DIST] + frac * (pos[DIST] - lastpos[DIST]))
  do i = 1, nex {
	extra[i] = (lastextra[i] + frac * (readextra[i] - lastextra[i]))
  }
end

