# Copyright (c) 1991 - 1900.  The Regents of the University of
# California (Regents).  All Rights Reserved.
# H+
# 	Title	:  t_periodint.x
# 	Author	:  tnk
# 	Date	:  05/19/93
# 	Synopsis:  
# 	Keywords:
# 		@(#) t_periodint.x -r1.1 05/19/93 UCB SSL CEA EGOCS
# 	Revisions:
# 	mm/dd/yy   name		description
# 	05/19/93   tnk	date and time created 93/05/19 12:22:28 by tnk
# H-
# 
#


# @(#)t_periodint.x	1.1 5/19/93 UCB CEA EGO
# AUTHOR : T. Kilsdonk
# DATE: 14 May 93

# INCLUDE FILES
include <tbset.h>
include <math.h>
include <time.h>

# periodint -- create a table of periodic time intervals
#
# Construct a set of time intervals that include
# or exclude a given phase of a periodically varying source,
# for the duration of an observation.
#
# Epoch is the time of some previous maximum point
# of the periodic cycle,
# i.e., when the period phase is zero.
# The units used are given in [epochunits], as
# JD (Julian date), MJD (modified Julian date),
# or euvesecs, where 
#
#         MJD = JD - 2400000.5
#    euvesecs = (MJD - 40000) * (24*60*60)
#
# MJD is used for all internal calculations of timestamps.
# Period length calculations are done in days.
#
# Results are written to either an ST table or an ASCII file
# ************************************************

# allowed time units
define JD_UNITS     "jd"
define MJD_UNITS    "mjd"
define EUVE_UNITS   "euvesecs"

# allowed units for period
define PERIOD_SECS  "seconds"
define PERIOD_MINS  "minutes"
define PERIOD_HOURS "hours"
define PERIOD_DAYS  "days"

# time conversion constants & macros
define MJD_OFFSET  2400000.5D0    #no. of days from JD to MJD
define EUVE_OFFSET   40000.0d0    #no. of days from MJD to EUVE epoch
define SECS_DAY      86400.0d0    #no. of seconds in a day
define MINS_DAY       1440.0d0    #no. of minutes in a day
define HOURS_DAY        24.0d0    #no. of hours in a day

define JD_TO_MJD       ( ($1) - MJD_OFFSET )
define MJD_TO_JD       ( ($1) + MJD_OFFSET )
define EUVESECS_TO_MJD ( ($1) / SECS_DAY + EUVE_OFFSET )
define MJD_TO_EUVESECS ( ( ($1) - EUVE_OFFSET ) * SECS_DAY )
define SECS_TO_DAYS    ( ($1) / SECS_DAY )
define MINS_TO_DAYS    ( ($1) / MINS_DAY )
define HOURS_TO_DAYS   ( ($1) / HOURS_DAY )

define TIME_FORMAT "%16.6f"    # output print format, ok for all units

procedure t_periodint()    #EXTERNAL

# parameter file strings
string OUTNAME_PARM      "output"      # str
string EPOCH_PARM        "epoch"       # double
string EP_UNITS_PARM     "epochunits"  # str
string PERIODLEN_PARM    "periodlen"   # double
string PER_UNITS_PARM    "periodunits" # str
string BEG_PHASE_PARM    "startphase"  # double
string END_PHASE_PARM    "endphase"    # double
string INCLUDE_PARM      "include"     # bool
string BEG_OBS_PARM      "startobs"    # double
string END_OBS_PARM      "endobs"      # double
string OBS_UNITS_PARM    "obsunits"    # str
string OUTTABLE_PARM     "table"       # bool

# LOCAL VARS
# parameters:
pointer out_name       # output file name (stdout is ok)
double  epoch          # period zero phase point
pointer epoch_units    # JD, MJD, euvesecs
double  period_len     # length of period
pointer period_units   # seconds|minutes|hours|days
double  beg_phase      # starting selection phase [0.0-1.0]
double  end_phase      # ending selection phase [0.0-1.0]
bool    include_it     # yes=keep, no=reject  startphase to endphase
double  beg_obs        # observation start time
double  end_obs        # observation end time
pointer obs_units      # JD, MJD, euvesecs
bool    do_table       # output format:  yes is STtable, no is ASCII
# 
pointer   stackp       # stack pointer 
double  beg_int_ph     # beginning phase of output time interval
double  end_int_ph     # ending phase of output time interval
double  beg_obs_ph     # beginning phase of observation
double  n_cycles       # no. of cycles between epoch & beg_obs
double  ph_adv         # phase advance
double  ph_diff        # phase difference
double  start_time     # beginning time of output time interval
double  end_time       # end of output time interval
double  int_len        # length of interval (time)
pointer out_units      # units to use for output values
int     out_fd         # output file descriptor (ASCII)
pointer out_tab        # output table ptr
pointer temp_str       # temp string
pointer stime_col      # ST table start time column pointer
pointer etime_col      # ST table end time column pointer
int     out_row        # output row number

# FUNCTIONS
double  time_to_mjd()  # convert various units to mjd
double  time_to_days() # convert various units to days
int     open()         # open file
pointer tbtopn()       # table initialization
bool    clgetb()       # get boolean CL parameter
double  clgetd()       # get double CL parameter
bool    streq()        # string compare

begin

  # mark the stack
  call smark(stackp)              

  # get parameters from param file
  call eu_clstr(out_name,     OUTNAME_PARM) 
  epoch       = clgetd(       EPOCH_PARM)
  call eu_clstr(epoch_units,  EP_UNITS_PARM) 
  period_len  = clgetd(       PERIODLEN_PARM)
  call eu_clstr(period_units, PER_UNITS_PARM) 
  beg_phase   = clgetd(       BEG_PHASE_PARM)
  end_phase   = clgetd(       END_PHASE_PARM)
  include_it  = clgetb(       INCLUDE_PARM) 
  beg_obs     = clgetd(       BEG_OBS_PARM)
  end_obs     = clgetd(       END_OBS_PARM)
  call eu_clstr(obs_units,    OBS_UNITS_PARM) 
  do_table    = clgetb(       OUTTABLE_PARM) 

  call salloc(temp_str, SZ_LINE, TY_CHAR)

  # open output file/table
  if (do_table) {
     # ST table
     out_tab = tbtopn(Memc[out_name], NEW_FILE, 0)
     call prep_table(out_tab,stime_col,etime_col)
  }
  else {
     # ASCII file

     # check for stdout (make case insensitive)
     call strcpy(Memc[out_name],Memc[temp_str],SZ_LINE)     
     call strlwr(Memc[temp_str])

     if (streq(Memc[temp_str],"stdout") )
        out_fd = STDOUT
     else {
        out_fd = open(Memc[out_name], NEW_FILE, TEXT_FILE)    
     }

  } 
  
  # convert all units to MJD 
  call strlwr(Memc[epoch_units])  # make lowercase
  epoch = time_to_mjd(epoch,Memc[epoch_units])

  call strlwr(Memc[obs_units])    # make lowercase
  beg_obs = time_to_mjd(beg_obs,Memc[obs_units])
  end_obs = time_to_mjd(end_obs,Memc[obs_units])

  # QC some of the parameters
  if (epoch > beg_obs) {
     call eprintf("%s must be <= %s\n")
     call pargstr(EPOCH_PARM)
     call pargstr(BEG_OBS_PARM)
     call error(110,"Exiting...")
  }

  if (period_len <= 0.0d0) {
     call eprintf("%s must be > zero\n")
     call pargstr(PERIODLEN_PARM)
     call error(120,"Exiting...")
  }

  if (beg_phase < 0.0d0 || beg_phase > 1.0d0) {
     call eprintf("%s must be > 0.0 and < 1.0\n")
     call pargstr(BEG_PHASE_PARM)
     call error(120,"Exiting...")
  }

  if (end_phase < 0.0d0 || end_phase > 1.0d0) {
     call eprintf("%s must be > 0.0 and < 1.0\n")
     call pargstr(END_PHASE_PARM)
     call error(120,"Exiting...")
  }

  # convert period to days
  call strlwr(Memc[period_units])  # make lowercase
  period_len = time_to_days(period_len,Memc[period_units])

  # use the observation units as the output units  
  call salloc(out_units, SZ_LINE, TY_CHAR)
  call strcpy(Memc[obs_units],Memc[out_units],SZ_LINE)

  # find beginning & ending phases of the [include/exclude] phase interval
  if (include_it) {          #include phase range
     beg_int_ph = beg_phase
     end_int_ph = end_phase
  }
  else {                     #exclude phase range
     beg_int_ph = end_phase
     end_int_ph = beg_phase     
  }

  # find time length of intervals
  ph_diff = (end_int_ph - beg_int_ph)
  if (ph_diff < 0.0d0) ph_diff = ph_diff + 1.0d0
  int_len = ph_diff * period_len

  # find the phase of the beginning observation time
  n_cycles = (beg_obs - epoch) / period_len
  beg_obs_ph = n_cycles - int(n_cycles)

  # find start time of desired first interval
  # (see how much time we have to advance from beg_obs)
  ph_adv = (beg_int_ph - beg_obs_ph)
  if (ph_adv < 0.0d0) ph_adv = ph_adv + 1.0d0
  start_time = beg_obs + (ph_adv * period_len)

  out_row = 1

  # see if we need to put out a partial first time interval
  end_time = start_time - (1.0d0 - ph_diff) * period_len
  if (end_time > beg_obs) {

     if (do_table) {
        call print_ST(out_tab,stime_col,beg_obs,etime_col,end_time,
           Memc[out_units],out_row)
     }
     else {
        call print_ascii(out_fd,beg_obs,end_time,Memc[out_units])    
     }

     out_row = 2
  }

  # now loop, putting out time segments til we hit end of obs time
  while (start_time < end_obs) {

     end_time = start_time + int_len
     if (end_time > end_obs) end_time=end_obs

     if (do_table) {
        call print_ST(out_tab,stime_col,start_time,etime_col,end_time,
           Memc[out_units],out_row)
     }
     else {
        call print_ascii(out_fd,start_time,end_time,Memc[out_units])
     }

     start_time = start_time + period_len
     out_row = out_row + 1
  }

  # close output file
  if (do_table)
     call tbtclo(out_tab)
  else
     call close(out_fd)

  # free the stack
  call sfree(stackp)


end
#***************************************************
# write 1 time interval to the output ascii file
procedure print_ascii(out_fd,start_time,end_time,out_units) #LOCAL
int out_fd             #I output file descriptor
double start_time      #I start time (MJD)
double end_time        #I end time (MJD)
char out_units[ARB]    #I output units

double start_conv      
double end_conv

double mjd_to_time()

begin

     start_conv = mjd_to_time(start_time,out_units)
     end_conv   = mjd_to_time(end_time,out_units)

     # print "starttime endtime\n"
     call fprintf(out_fd,TIME_FORMAT)
     call pargd(start_conv)
     call fprintf(out_fd," ")
     call fprintf(out_fd,TIME_FORMAT)
     call pargd(end_conv)
     call fprintf(out_fd,"\n")


end
#***************************************************
# write 1 time interval to the output ST table
procedure print_ST(out_tab,stime_col,start_time,etime_col,end_time,
   out_units,out_row)  #LOCAL
pointer out_tab        #I output file descriptor
pointer stime_col      #I start time column pointer
double start_time      #I start time (MJD)
pointer etime_col      #I end time column pointer
double end_time        #I end time (MJD)
char out_units[ARB]    #I output units
int out_row            #I output row

double start_conv      
double end_conv

double mjd_to_time()

begin

     start_conv = mjd_to_time(start_time,out_units)
     end_conv   = mjd_to_time(end_time,out_units)

     call tbrptd(out_tab, stime_col, start_conv, 1, out_row)
     call tbrptd(out_tab, etime_col, end_conv, 1, out_row)

end
#****************************************
# convert from JD,MJD,euvesecs to MJD
double procedure time_to_mjd(in_time,units)   #LOCAL
double in_time      #I input time in "units"
char   units[ARB]   #I units to convert from

double mjd
bool streq()

begin

  if ( streq(units,JD_UNITS) ) {
     mjd = JD_TO_MJD(in_time)
  }
  else if ( streq(units,EUVE_UNITS) ){
     mjd = EUVESECS_TO_MJD(in_time)
  }
  else if ( streq(units,MJD_UNITS) ) {
     mjd = in_time
  }
  else {
     call eprintf("Invalid time units %s\n")
     call pargstr(units)
     call error(100,"Aborting")
  }

return mjd

end
#****************************************
# convert from secs,mins,hours to days
double procedure time_to_days(in_time,units)  #LOCAL
double in_time      #I input time in "units"
char   units[ARB]   #I units to convert from

double days
bool streq()

begin

  if ( streq(units,PERIOD_SECS) ) {
     days = SECS_TO_DAYS(in_time)
  }
  else if ( streq(units,PERIOD_MINS) ){
     days = MINS_TO_DAYS(in_time)
  }
  else if ( streq(units,PERIOD_HOURS) ){
     days = HOURS_TO_DAYS(in_time)
  }
  else if ( streq(units,PERIOD_DAYS) ){
     days = in_time
  }
  else {
     call eprintf("Invalid time units %s\n")
     call pargstr(units)
     call error(100,"Aborting")
  }

return days

end
#****************************************
# convert MJD to other time units 
double procedure mjd_to_time(in_time,units)  #LOCAL
double in_time      #I input time in mjd
char   units[ARB]   #I units to convert to

double time
bool streq()

begin

  if ( streq(units,JD_UNITS) ) {
     time = MJD_TO_JD(in_time)
  }
  else if ( streq(units,EUVE_UNITS) ){
     time = MJD_TO_EUVESECS(in_time)
  }
  else if ( streq(units,MJD_UNITS) ){
     time = in_time
  }
  else {
     call eprintf("Invalid time units %s\n")
     call pargstr(units)
     call error(100,"Aborting")
  }

return time

end
#*************************************
# set up the output ST table 
procedure prep_table(out_tab,stime_col,etime_col)     #LOCAL
pointer out_tab      #I output table pointer (opened)
pointer stime_col    #O start time column pointer
pointer etime_col    #O end time column pointer

define START_COL_NAME  "stime"
define END_COL_NAME    "etime"

begin


  #define the output columns
  #                   --->  Name        Unit       Format
  call tbcdef(out_tab, stime_col, START_COL_NAME, 
                       "Seconds", TIME_FORMAT, TY_DOUBLE,1,1)

  call tbcdef(out_tab, etime_col, END_COL_NAME, 
                       "Seconds", TIME_FORMAT, TY_DOUBLE,1,1)

  # create the table
  call tbtcre(out_tab)

end


