# SPP/Yacc specification for a quaternion calculator.

%{
include <iraf.h>

include	<ctype.h>
include	<lexnum.h>

include	<fset.h>
include	<math.h>

define	YYMAXDEPTH	150		# length of parser stack

#task	qcalc = t_qcalc

# Operand Structure (parser stack)
define	YYOPLEN		10		# size of operand structure
define	OPTYPE		Meml[P2L($1)+0]	# Operand type

define	OPVALI		Meml[P2L($1)+1]	# letter

define	OPVALW		Memd[P2D($1)+1]	# W component
define	OPVALX		Memd[P2D($1)+2]	# X component
define	OPVALY		Memd[P2D($1)+3]	# Y component
define	OPVALZ		Memd[P2D($1)+4]	# Z component

%}

%token	CONST LETTER YYEOF

%left	'!' '&' '^'
%left	'?' ':'
%left	'+' '-'
%left	'*' '/'
%left	UMINUS
%left	'@' '%'

%%

prog	:	# Empty
	|	prog stmt eost {
			return (OK)
		    }
	|	YYEOF {
			return (EOF)
		    }
	|	prog error '\n' {
			yyerrok
		    }
	|	prog error YYEOF {
			return (EOF)
		    }
	;

stmt	:	expr {
			# Print the value of an expression.
    call printf("%g %g %g %g\n")
    call pargd(OPVALW($1))
    call pargd(OPVALX($1))
    call pargd(OPVALY($1))
    call pargd(OPVALZ($1))
    call flush(STDOUT)
    }
stmt	:	'&' opnl expr {
			# Print the value of an expression as vector
    call printf("%g %g %g\n")
    call pargd(OPVALX($3))
    call pargd(OPVALY($3))
    call pargd(OPVALZ($3))
    call flush(STDOUT)
    }
stmt	:	'!' opnl expr {
			# Print the value of an expression as position
    call qc_position($3, 0)
    call flush(STDOUT)
    }
stmt	:	'^' opnl expr {
			# Print the value of an expression as position
    call qc_position($3, 1)
    call flush(STDOUT)
    }
	|	LETTER '=' expr {
			# Set the value of a register (from a-z).
			call putreg (OPVALI($1), $3)
		    }
	;

expr	:	'(' opnl expr opnl ')' {
			YYMOVE ($3, $$)
		    }
	|	expr '+' opnl expr {
			call binop ($1, $4, $$, '+')
		    }
	|	expr '-' opnl expr {
			call binop ($1, $4, $$, '-')
		    }
	|	expr '*' opnl expr {
			call binop ($1, $4, $$, '*')
		    }
	|	expr '@' opnl expr {
			call binop ($1, $4, $$, '@')
		    }
	|	expr '%' opnl expr {
			call binop ($1, $4, $$, '%')
		    }
	|	expr '/' opnl expr {
			call binop ($1, $4, $$, '/')
		    }
	|	'-' expr %prec UMINUS {
			call unop ($2, $$, '-')
		    }
        |       expr '?' opnl expr opnl ':' opnl expr {
			if (OPVALW($1) > 0)	YYMOVE ($4, $$)
			else			YYMOVE ($8, $$)
                    }
        |       '|' expr opnl '|' {
                        call unop($3, $$, '|')
                    }
	|	LETTER {
			call getreg (OPVALI($1), $$)
		    }
	|	CONST
        |       '<' opnl expr opnl '>' {
                        OPVALW($$) = 0
			call qc_decdeg($3,$$)
                    }
        |       '{' opnl expr opnl '}' {
                        OPVALW($$) = 0
			call qc_dechour($3,$$)
                    }
	;

eost	:	';'
	|	'\n'
	;

opnl	:	# Empty
	|	opnl '\n'
	;

%%


# QCALC -- Main routine for the quaternion calculator.

procedure t_qcalc()

int	status
int	yyparse()
int	yylex()
extern	yylex()
bool	debug
bool	clgetb()
pointer	sp, in
int	strlen()
int	open()
int	stropen()
bool	streq()
bool	filep
int	fd
int	len

begin
  call smark(sp)
  call salloc(in, SZ_LINE, TY_CHAR)
  debug = FALSE
  call clgstr("input", Memc[in], SZ_LINE)
  filep = clgetb("file")
  if (filep) {
    if (streq(Memc[in], "STDIN")) {
      fd = open(Memc[in], READ_ONLY, TEXT_FILE)
    } else {
      fd = STDIN
    }
  } else {
    len = strlen(Memc[in])
    if (len < 1 || Memc[in+len-1] != ';') {
      Memc[in+len] = ';'
      len = len + 1
      Memc[in+len] = EOS
    }
    fd = stropen(Memc[in], strlen(Memc[in]), READ_ONLY)
  }

  repeat {
    status = yyparse (fd, debug, yylex)
    if (status == ERR)
      call eprintf ("syntax error")
    } until (status == EOF)

  if (filep) {
    if (fd != STDIN) call close(fd)
  } else {
    call strclose(fd)
  }

  call sfree(sp)
end

procedure binop (a, b, c, operation)

pointer	a, b, c			# c = a op b
int	operation		# i.e., '*', '/', etc.
double	qa[4], qb[4], qc[4]

begin
  qa[1] = OPVALW(a)
  qa[2] = OPVALX(a)
  qa[3] = OPVALY(a)
  qa[4] = OPVALZ(a)

  qb[1] = OPVALW(b)
  qb[2] = OPVALX(b)
  qb[3] = OPVALY(b)
  qb[4] = OPVALZ(b)

  switch (operation) {
    case '+':
      qc[1] = qa[1] + qb[1]
      qc[2] = qa[2] + qb[2]
      qc[3] = qa[3] + qb[3]
      qc[4] = qa[4] + qb[4]
    case '-':
      qc[1] = qa[1] - qb[1]
      qc[2] = qa[2] - qb[2]
      qc[3] = qa[3] - qb[3]
      qc[4] = qa[4] - qb[4]
    case '*':
      call er_qmult(qa, qb, qc)
    case '%':
      call er_qmkrotate(qc, qa, qb)
    case '@':
      qc[1] = qb[1]
      qc[2] = qb[2]
      qc[3] = qb[3]
      qc[4] = qb[4]
      call er_qrotateq(qc, qa)
    case '/':
      call er_qinvert(qb, qb)
      call er_qmult(qa, qb, qc)
    default:
      call error (1, "unknown binary operator")
  }

  OPVALW(c) = qc[1]
  OPVALX(c) = qc[2]
  OPVALY(c) = qc[3]
  OPVALZ(c) = qc[4]
end


procedure unop (a, b, operation)

pointer	a, b
int	operation
double	qa[4], qb[4], qt[4]
double	sqrt()

begin
  qa[1] = OPVALW(a)
  qa[2] = OPVALX(a)
  qa[3] = OPVALY(a)
  qa[4] = OPVALZ(a)

	switch (operation) {
	case '-':
	  call er_qinvert(qa, qb)
	case '|':
	  call er_qinvert(qa, qt)
	  call er_qmult(qa, qt, qb)
	  qb[1] = sqrt(qb[1])
	default:
	  call error (2, "unknown unary operator")
	}

  OPVALW(b) = qb[1]
  OPVALX(b) = qb[2]
  OPVALY(b) = qb[3]
  OPVALZ(b) = qb[4]
end


define	MAXREG		('z'-'a'+1)

procedure getreg (regchar, op)

int	regchar
pointer	op

bool	store
real	regbuf[MAXREG*YYOPLEN]
int	reg, offset

begin
	store = false
	goto 10

entry	putreg (regchar, op)
	store = true

	# Compute offset into storage.  Structures are stored in buffer
	# by a binary copy, knowing only the length of the structure.
10	if (IS_UPPER(regchar))
	    reg = regchar - 'A' + 1
	else
	    reg = regchar - 'a' + 1
	reg = max(1, min(MAXREG, reg))
	offset = (reg-1) * YYOPLEN + 1

	# Copy the operand structure either in or out.
	if (store)
	    call amovr (Memr[op], regbuf[offset], YYOPLEN)
	else
	    call amovr (regbuf[offset], Memr[op], YYOPLEN)
end


procedure	qc_dechour (a, b)
pointer		a, b

double	ra, dec
double	sin(), cos()
begin
  ra = OPVALW(a) * PI / 12
  dec = OPVALX(a) * PI / 180
  OPVALW(b) = 0
  OPVALX(b) = cos(ra)*cos(dec)
  OPVALY(b) = sin(ra)*cos(dec)
  OPVALZ(b) = sin(dec)
end

procedure	qc_decdeg (a, b)
pointer		a, b

double	ra, dec
double	sin(), cos()
begin
  ra = OPVALW(a) * PI / 180
  dec = OPVALX(a) * PI / 180
  OPVALW(b) = 0
  OPVALX(b) = cos(ra)*cos(dec)
  OPVALY(b) = sin(ra)*cos(dec)
  OPVALZ(b) = sin(dec)
end

procedure	qc_position (a, form)
pointer	a
int	form

double	ra, dec
double	atan2(), sqrt()

begin
    if (OPVALX(a) == 0 && OPVALY(a) == 0)
      ra = 0
    else
      ra = atan2(OPVALY(a), OPVALX(a))
    if (ra < 0) ra = ra + TWOPI

    dec = atan2(OPVALZ(a), sqrt(OPVALX(a)**2+OPVALY(a)**2))

    if (form == 0) {
      call printf("%g %g\n")
      call pargd(ra * 180 / PI)
      call pargd(dec * 180 / PI)
    } else {
      call printf("%h %h\n")
      call pargd(ra * 12 / PI)
      call pargd(dec * 180 / PI)
    }
end

# YYLEX -- Lexical input routine.  Return next token from the input
# stream.  Recognized tokens are CONST (numeric constants), LETTER,
# and the operator characters.

int procedure yylex (fd, yylval)

int	fd
pointer	yylval
char	ch, lbuf[SZ_LINE]
int	ip, junk
double	dval
int	getline(), gctod()
data	ip /0/
int	fstati()

begin
	# Fetch a nonempty input line, or advance to start of next token
	# if within a line.  Newline is a token.
	repeat {
	    if (ip <= 0 || lbuf[ip] == EOS) {
	        if (fd == STDIN && fstati(fd, F_REDIR) == NO	&&
		    fstati(STDOUT, F_REDIR) == NO) {
		  call printf("qcalc> ")
		  call flush(STDOUT)
		}
		if (getline (fd, lbuf) == EOF) {
		  if (fd == STDIN && fstati(fd, F_REDIR) == NO	&&
		      fstati(STDOUT, F_REDIR) == NO) {
		    call printf("\n")
		    call flush(STDOUT)
		  }
		  ip = 0
		  return (YYEOF)
		} else
		    ip = 1
	    }
	    while (IS_WHITE (lbuf[ip]))
		ip = ip + 1
	} until (lbuf[ip] != EOS)

	# Determine type of token.  If numeric constant, convert to binary
	# and return value in op structure (yylval).  If letter (register
	# variable) return value and advance input one char.  If any other
	# character, return char itself as the token, and advance input one
	# character.

	if ( !(IS_DIGIT(lbuf[ip]) || (lbuf[ip] == '.')) ) {
	    ch = lbuf[ip]
	    ip = ip + 1
	    if (IS_ALPHA (ch)) {
		OPTYPE(yylval) = LETTER
		OPVALI(yylval) = ch
		return (LETTER)
	    } else {
		OPTYPE(yylval) = ch
		return (OPTYPE(yylval))
	    }
	} else {
	    OPTYPE(yylval) = CONST
	    junk = gctod (lbuf, ip, dval)
	    OPVALW(yylval) = dval

	    junk = gctod (lbuf, ip, dval)
	    if (junk != 0)	OPVALX(yylval) = dval
	    else		OPVALX(yylval) = 0

	    junk = gctod (lbuf, ip, dval)
	    if (junk != 0)	OPVALY(yylval) = dval
	    else		OPVALY(yylval) = 0

	    junk = gctod (lbuf, ip, dval)
	    if (junk != 0)	OPVALZ(yylval) = dval
	    else		OPVALZ(yylval) = 0


	    return (CONST)
        }
end
